# Generated by Django 3.0.8 on 2020-09-08 10:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0020_auto_20200907_0753'),
    ]

    operations = [
        migrations.AddField(
            model_name='tag',
            name='description',
            field=models.TextField(blank=True, max_length=1024, null=True),
        ),
    ]



import logging

logger = logging.getLogger(__name__)


def get_fields(model):
    """
    Returns django fields of current ``model``.

    Does not include inherited fields.
    """
    return model._meta.get_fields(include_parents=False)


def group_per_model(models, **kwargs):
    """
    groups kwargs per model

    What this method is supposed to do is better exmplained by example.
    Suppose there are 3 models i.e. ``models`` list contains
    3 models: [Model_1, Model_2, Model_3]. Each of these models
    has following django field attributes:

        class Model_1:
            attr_m1_a
            attr_m1_b

        class Model_2:
            attr_m2_x

        class Model_3:
            attr_m3_j
            attr_m3_k

    kwargs_ex_1 = {
        'blah': 1, 'extra': 2, 'attr_m1_a': "see?", 'attr_m1_b': "this?"
    }

    group_per_model(models, **kwargs_ex_1) will return:

        {
            Model_1: {'attr_m1_a': "see?", 'attr_m1_b': "this?"}
        }

    kwargs_ex_2 = {
        'blah': 1, attr_m1_a': "AB", 'attr_m2_x': "XX"
    }
    group_per_model(models, **kwargs_ex_2) will return:

        {
            Model_1: {'attr_m1_a': "AB"}
            Model_2: {'attr_m2_x': "XX"}
        }
    """
    ret = {}

    for model in models:
        fields = get_fields(model)
        for field in fields:
            if field.name in kwargs.keys():
                ret.setdefault(model, {}).update(
                    {field.name: kwargs[field.name]}
                )

    return ret



from rest_framework import serializers

from papermerge.core.models import (Document, User)


class UserSerializer(serializers.ModelSerializer):

    documents = serializers.PrimaryKeyRelatedField(
        many=True, queryset=Document.objects.all()
    )

    class Meta:
        model = User
        fields = ['id', 'username', 'documents']


class DocumentSerializer(serializers.Serializer):
    id = serializers.IntegerField(read_only=True)
    title = serializers.CharField(
        required=False,
        allow_blank=True,
        max_length=100
    )
    notes = serializers.CharField(
        required=False,
        allow_blank=True,
        allow_null=True,
        max_length=100
    )
    page_count = serializers.IntegerField(
        read_only=True
    )

    user = serializers.ReadOnlyField(source='user.username')

    def update(self, instance, validated_data):
        """
        Update and return an existing `Document` instance,
        given the validated data.
        """
        instance.title = validated_data.get('title', instance.title)
        instance.notes = validated_data.get('notes', instance.notes)
        instance.save()

        return instance

    def create(self, validated_data):
        """
        Create and return a new `Document` instance, given the validated data.
        """
        return Document.objects.create(**validated_data)



import logging

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext as _


class LogEntry(models.Model):
    """
    LogEntry visible on left side panel.
    Each user can see his/her logs.
    Superuser can see everybody's log.

    LogEntries are very useful to gets insights into OCRs and Automates
    activity. In general to see what is happening in the Papermerge system.
    """

    LEVELS = (
        (logging.DEBUG, _("Debug")),
        (logging.INFO, _("Info")),
        (logging.WARNING, _("Warning")),
        (logging.ERROR, _("Error")),
        (logging.CRITICAL, _("Critical")),
    )

    message = models.TextField()
    level = models.PositiveIntegerField(
        choices=LEVELS,
        default=logging.INFO
    )

    action_time = models.DateTimeField(
        _('action time'),
        default=timezone.now,
        editable=False,
    )

    user = models.ForeignKey(
        'core.User',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    class Meta:
        verbose_name = _('log entry')
        verbose_name_plural = _('log entries')
        ordering = ('-action_time',)



from django.urls import include, path
from .views import access as access_views
from .views import tags as tags_views
from .views import api as api_views
from .views import documents as doc_views
from .views import nodes as node_views
from .views import metadata as metadata_views
from .views import users as users_views
from .views import automates as automate_views

document_patterns = [
    path(
        '<int:doc_id>/',
        doc_views.document,
        name="document"
    ),
    path(
        '<int:id>/preview/page/<int:page>',
        doc_views.preview,
        name="preview"
    ),
    path(
        '<int:id>/preview/<int:step>/page/<int:page>',
        doc_views.preview,
        name="preview"
    ),
    path(
        '<int:id>/hocr/<int:step>/page/<int:page>',
        doc_views.hocr,
        name="hocr"
    ),
    path(
        '<int:id>/<int:document_version>/text/page/<int:page_number>',
        doc_views.text_view,
        name="text_view"
    ),
    path(
        'usersettings/<str:option>/<str:value>',
        doc_views.usersettings,
        name="usersettings"
    ),
]

app_name = 'core'

urlpatterns = [
    path(
        'document/', include(document_patterns)
    ),
    path('browse/', node_views.browse_view, name="browse"),
    path('browse/<int:parent_id>/', node_views.browse_view, name="browse"),
    path('breadcrumb/', node_views.breadcrumb_view, name="breadcrumb"),
    path(
        'breadcrumb/<int:parent_id>/',
        node_views.breadcrumb_view,
        name="breadcrumb"
    ),
    path('node/<int:node_id>', node_views.node_view, name="node"),
    # Node can be a document or a folder
    # Downloading entire folder by selecting it - makes perfect sense
    path(
        'node/<int:id>/download/',
        node_views.node_download,
        name="node_download"
    ),
    path(
        'node/by/title/<str:title>', node_views.node_by_title_view,
        name="node_by_title"
    ),
    path('nodes/', node_views.nodes_view, name="nodes"),
    # when user selected multiple documents and folders for download
    path('nodes/download/', node_views.nodes_download, name="nodes_download"),
    path(
        'node/<int:id>/access', access_views.access_view, name="access"
    ),
    path(
        'node/<int:node_id>/tags/', tags_views.tags_view, name="tags"
    ),
    path(
        'nodes/tags/', tags_views.nodes_tags_view, name="nodes_tags"
    ),
    path(
        'alltags/', tags_views.alltags_view, name="alltags"
    ),
    path(
        'metadata/<model>/<int:id>', metadata_views.metadata, name="metadata"
    ),
    path(
        'usergroups', access_views.user_or_groups, name="user_or_groups"
    ),
    path(
        'upload/',
        doc_views.upload,
        name="upload"
    ),
    path(
        'create-folder/',
        doc_views.create_folder,
        name='create_folder'
    ),
    path(
        'rename-node/<int:id>',
        doc_views.rename_node,
        name='rename_node'
    ),
    path(
        'cut-node/',
        doc_views.cut_node,
        name='cut_node'
    ),
    path(
        'paste-node/',
        doc_views.paste_node,
        name='paste_node'
    ),
    path(
        'paste-pages/',
        doc_views.paste_pages,
        name='paste_pages'
    ),
    path(
        'clipboard/',
        doc_views.clipboard,
        name='clipboard'
    ),
    path(
        'run-ocr/', doc_views.run_ocr_view, name="run_ocr"
    ),
    path(
        'api/documents',
        api_views.DocumentsView.as_view(),
        name='api_documents'
    ),
    path(
        'api/document/upload/<str:filename>',
        api_views.DocumentUploadView.as_view(),
        name='api_document_upload'
    ),
    path(
        'api/document/<int:pk>/',
        api_views.DocumentView.as_view(),
        name='api_document'
    ),
    path(
        'api/document/<int:doc_id>/pages',
        api_views.PagesView.as_view(),
        name='api_pages'
    ),
    path(
        'api/document/<int:doc_id>/pages/cut',
        api_views.PagesCutView.as_view(),
        name='api_pages_cut'
    ),
    path(
        'api/document/<int:doc_id>/pages/paste',
        api_views.PagesPasteView.as_view(),
        name='api_pages_paste'
    ),
    path(
        'user/<int:id>/change-password',
        users_views.user_change_password_view, name='user_change_password'
    ),
    path(
        'automate/',
        automate_views.automate_view,
        name="automate_view"
    ),
    path(
        'automate/<int:automate_id>/',
        automate_views.automate_change_view, name="automate_change_view"
    )
]



from django.urls import reverse


class PreviewUrlsHandover:
    """
    Generate all preview urls for given document id.

    Generated urls might be passed to static js code in order
    to enpower javasscript code the initiation of previews.
    """

    def __init__(
        self,
        document_id,
        page_count
    ):
        self.document_id = document_id
        self.page_count = page_count

    def __iter__(self):
        for index in range(1, self.page_count + 1):
            doc_id = self.document_id
            for step in range(0, 4):
                yield(
                    (
                        index,
                        step,
                        reverse('core:preview', args=[doc_id, step, index]))
                )



import logging

from django.contrib.auth.decorators import login_required
from django.shortcuts import get_object_or_404

from papermerge.core.models import Automate, Tag
from .decorators import json_response

logger = logging.getLogger(__name__)


@json_response
@login_required
def automate_view(request):
    """
    If no particular automate is asked for, just return
    list of all tags (which might be associated to the new automate)
    """
    tags = Tag.objects.filter(user=request.user)

    response = {
        'tags': [],
        'alltags': [tag.to_dict() for tag in tags]
    }

    return response


@json_response
@login_required
def automate_change_view(request, automate_id):
    """
    Returns details of the automate specified with given id
    """
    automate = get_object_or_404(Automate, id=automate_id)
    tags = automate.tags.all()
    alltags = Tag.objects.filter(user=request.user)

    response = {
        'name': automate.name,
        'id': automate_id,
        'tags': [tag.to_dict() for tag in tags],
        'alltags': [tag.to_dict() for tag in alltags],
    }

    return response



# Generated by Django 3.1.3 on 2020-12-01 11:16

from django.db import migrations, models
import papermerge.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0033_auto_20201119_0032'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='document',
            name='file_orig',
        ),
        migrations.AlterField(
            model_name='basetreenode',
            name='title',
            field=models.CharField(max_length=200, validators=[papermerge.core.validators.safe_character_validator], verbose_name='Title'),
        ),
        migrations.AlterField(
            model_name='document',
            name='file_name',
            field=models.CharField(default='', max_length=1024, validators=[papermerge.core.validators.safe_character_validator]),
        ),
        migrations.AlterField(
            model_name='document',
            name='notes',
            field=models.TextField(blank=True, null=True, validators=[papermerge.core.validators.safe_character_validator], verbose_name='Notes'),
        ),
        migrations.AlterField(
            model_name='document',
            name='text',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(max_length=100, validators=[papermerge.core.validators.safe_character_validator], verbose_name='name'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ValidationError
from django.utils.text import slugify

from papermerge.contrib.admin.forms import TagForm
from papermerge.core.models import Tag
from papermerge.contrib.admin.views import mixins as mix


class TagsView(LoginRequiredMixin):
    model = Tag
    form_class = TagForm
    success_url = reverse_lazy('admin:tags')


class TagsListView(
    TagsView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):

    title = _("Tags")

    def get_queryset(self):
        qs = super().get_queryset()

        return qs.filter(user=self.request.user)


class TagCreateView(TagsView, generic.CreateView):

    title = _('New Tag')

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('New')
        context['action_url'] = reverse_lazy('admin:tag-add')

        return context

    def form_valid(self, form):
        # set fields which user does not have access to
        form.instance.user = self.request.user
        # otherwise 'slug' cannot be empty Exception is rised
        form.instance.slug = slugify(form.cleaned_data['name'])
        try:
            form.instance.full_clean()
        except ValidationError as e:
            form._errors['name'] = [
                str(e)
            ]
            del form.cleaned_data["name"]

            return super().form_invalid(form)

        return super().form_valid(form)


class TagUpdateView(TagsView, generic.UpdateView):

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = 'Edit'
        context['action_url'] = reverse_lazy(
            'admin:tag-update',
            args=(self.object.pk,)
        )
        return context



# Generated by Django 3.0.2 on 2020-01-16 10:50
from django.db import migrations
#from papermerge.core.utils import get_sql_content


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]
    # DB dependent oprations moved out as plugin
    operations = [
#       migrations.RunSQL(
#           get_sql_content('01_triggers.sql')
#       ),
#       migrations.RunSQL(
#           get_sql_content('02_basetreenode.sql')
#       ),
#       migrations.RunSQL(
#           get_sql_content('03_update_lang_cols.sql')
#       ),
#       migrations.RunSQL(
#           get_sql_content('04_views.sql')
#       ),
    ]





class Diff:
    """
    A list of Models which can be propagated from parent to child node
    an operation associated.

    This structure is necessary because of
    the access permissions/kv/kvcomp
    propagation from parent to child nodes i.e. if some access/kv/kvcomp(es)
    is (are) applied to a node - it will affect (update, insert, delete)
    all its descendents.
    """
    DELETE = -1
    UPDATE = 0
    ADD = 1  # accesses in the list will be added
    REPLACE = 2

    def __init__(self, operation, instances_set=[]):
        self._op = operation

        if len(instances_set) == 0:
            self._set = set()
        else:
            self._set = instances_set

    @property
    def operation(self):
        return self._op

    def add(self, instance):
        self._set.add(instance)

    def __len__(self):
        return len(self._set)

    def __iter__(self):
        return iter(self._set)

    def first(self):
        if len(self) > 0:
            return list(self._set)[0]

        return None

    def pop(self):
        return self._set.pop()

    def is_update(self):
        return self.operation == self.UPDATE

    def is_add(self):
        return self.operation == self.ADD

    def is_delete(self):
        return self.operation == self.DELETE

    def is_replace(self):
        return self.operation == self.REPLACE

    def __str__(self):
        op_name = {
            self.DELETE: "delete",
            self.UPDATE: "update",
            self.ADD: "add",
            self.REPLACE: "replace"
        }
        inst_list = [
            str(inst) for inst in self._set
        ]
        op = op_name[self._op]

        return f"Diff({op}, {inst_list})"

    def __repr__(self):
        return self.__str__()



from django.conf import settings

"""
Tesseract uses ISO-639-2/T for language names
Postgres uses different way to name its languages.
LANG_DICT maps those two worlds.
"""
# key is ISO-639-2/T as per:
# https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
# value is language dictionary as as per
# psql> \dFd (which lists installed dictionaries)

LANG_DICT = settings.PAPERMERGE_OCR_LANGUAGES


def get_default_ocr_lang():

    return settings.PAPERMERGE_OCR_DEFAULT_LANGUAGE


def get_ocr_lang_choices(capitalize=True):
    """
    returns a list of tuples as required by
    Django's choices ((key, value),(key, value), ...)
    """

    return [
        (key, value.capitalize())
        for key, value in LANG_DICT.items()
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models

from papermerge.core.models import Access, Diff

# custom user is used - papermerge.core.models.User
User = get_user_model()


def delete_access_perms(node, access_list):
    """
    same input as for set_access_perms
    """

    access_diff_d = Diff(operation=Diff.DELETE)
    # by the user
    for access_item in node.access_set.all():
        if access_item.user:
            name = access_item.user.username
            model_type = 'user'
        elif access_item.group:
            name = access_item.group.name
            model_type = 'group'

        # if node's access list contains items
        # which are in access_list - delete respective
        # node's access
        filtered = [
            x for x in access_list
            if x['model'] == model_type and x['name'] == name
        ]

        if len(filtered) > 0:
            # further propagation is done via pre_delete
            # signals for Access models
            access_diff_d.add(access_item)
            access_item.delete()

    return [access_diff_d]


def set_access_perms(node, access_list):
    """
    node is an instance of core.models.BaseTreeNode
    access is an array/list of hashes of following
    format:

    access = [
        {
            'model': 'user',
            'access_type': 'allow',
            'name': 'margaret',
            'permissions': {
                'read': true,
                'write': true,
                'delete': true,
                'change_perm': true,
                'take_ownership': true
            }
        },
        {
            'model': 'group',
            'access_type': 'allow',
            'name': 'employee',
            'permissions': {
                'read': true,
                'write': true,
                'delete': true,
                'change_perm': true,
                'take_ownership': true
            }
        }, ...
    ]
    """
    # first, add new access entries to the node
    # or update existing onces.
    access_diff_u = Diff(operation=Diff.UPDATE)
    access_diff_a = Diff(operation=Diff.ADD)

    for access_hash in access_list:
        access = get_access_for(
            node=node,
            model_type=access_hash['model'],
            name=access_hash['name']
        )
        if access:
            if access.perm_diff(access_hash['permissions']):
                access.set_perms(access_hash['permissions'])
                access_diff_u.add(access)
        else:
            access = create_access(
                node=node,
                model_type=access_hash['model'],
                name=access_hash['name'],
                access_type=access_hash['access_type'],
                access_inherited=False,
                permissions=access_hash['permissions']
            )
            access_diff_a.add(access)

    ret = []
    if len(access_diff_a) > 0:
        ret.append(access_diff_a)

    if len(access_diff_u) > 0:
        ret.append(access_diff_u)

    return ret


def create_access(
    node,
    model_type,
    access_type,
    name,
    permissions,
    access_inherited
):
    """
    permissions is a dictionary of following format:

        {
            'read': True|False,
            'delete': True|False,
            'write': True|False,
            'take_ownership': True|False,
            'change_perm': True|False,
        }
    """
    content_type, _ = ContentType.objects.get_or_create(
        app_label="core",
        model="access",
    )
    # if key is either missing in permissions dictionary
    # (is not one of 5 permissions) or will have False value
    # permission wont be set.
    perm_keys = [  # permissions keys to set
        key for key, value in permissions.items() if value
    ]

    if len(perm_keys) == 0:
        raise Exception(
            "auth.create_access: I refuse to create access with "
            "empty permissions."
        )

    perms = Permission.objects.filter(
        codename__in=perm_keys,
        content_type=content_type
    )

    if perms.count() == 0 and len(perm_keys) > 0:
        raise Exception(
            f"Not a single of Core Access Permissions was "
            f"found: {perm_keys}."
        )

    if model_type == 'user':
        user = User.objects.get(username=name)
        access = Access.objects.create(
            user=user,
            access_type=access_type,
            node=node,
            access_inherited=access_inherited
        )
    elif model_type == 'group':
        group = Group.objects.get(name=name)
        access = Access.objects.create(
            group=group,
            access_type=access_type,
            node=node
        )

    access.permissions.add(*perms)

    return access


def get_access_perms_as_hash(
    node,
    model_type,
    name
):
    """
        node = core.models.BaseTreeNode instance
        model_type = 'user' or 'group' as string
        name = username in case of user, and name
            in case of group model type
    """
    access = get_access_for(node, model_type, name)

    if not access:
        return {}

    result = {}
    result[Access.PERM_READ] = access.has_perm(Access.PERM_READ)
    result[Access.PERM_WRITE] = access.has_perm(Access.PERM_WRITE)
    result[Access.PERM_DELETE] = access.has_perm(Access.PERM_DELETE)
    result[Access.PERM_CHANGE_PERM] = access.has_perm(Access.PERM_CHANGE_PERM)
    result[Access.PERM_TAKE_OWNERSHIP] = access.has_perm(
        Access.PERM_TAKE_OWNERSHIP
    )

    return result


def create_access_perms():
    """
    Permissions related to the Access model.
    Access model grants permissions to BaseTreeNodes.
    """
    results = []
    # make sure description here matches the one
    # of the papermerge.core.models.Access.Meta.permissions
    # i.e. Change Permissions Access, etc
    permissions = (
        (Access.PERM_CHANGE_PERM, "Change Permissions Access"),
        (Access.PERM_TAKE_OWNERSHIP, "Take Ownership Access"),
        (Access.PERM_READ, "Read Access"),
        (Access.PERM_WRITE, "Write Access"),
        (Access.PERM_DELETE, "Delete Access"),
    )
    content_type, _ = ContentType.objects.get_or_create(
        app_label="core",
        model="access",
    )
    for perm in permissions:
        p, dummy = Permission.objects.get_or_create(
            codename=perm[0],
            name=perm[1],
            content_type=content_type
        )
        dummy = dummy
        results.append(p)

    return results


def get_access_for(node, model_type, name):

    access = None

    if model_type == Access.MODEL_USER:
        access = Access.objects.filter(
            node=node,
            user__username=name
        ).first()
    elif model_type == Access.MODEL_GROUP:
        access = Access.objects.filter(
            node=node,
            group__name=name
        ).first()

    return access


class NodeAuthBackend:
    """
    In this context Object = Node = papermerge.core.models.BaseTreeNode

    Each object has many access model associated. Access model is either
    assigned directly to the node (object) or inherited via object's parent.
    Inheritance of access models happens during node creation or via user's
    management of access instances. In other words when deciding if user
    has permission for a node - node's parents are not consulted.

    Each Access model has one access type associated (ALLOW or DENY) and
    either one group or user, access model basically says:
        allow (or deny) access to this user (or group) using given permission
        set.

    In case there are multiple access instances with same permission but with
    different access_type (elizabeth is both allowed and denied read
    permission on node X) then deny access type has priority (elizabeth will
    be denied access to node X).
    """

    def authenticate(self, request, username=None, password=None):
        # mandatory interface for django auth backend.
        # But we don't use it.
        return None

    def get_user(self, user_id):
        # mandatory interface for django auth backend.
        # But we don't use it.
        return None

    def django_get_user_permissions(self, user_obj):
        return user_obj.role.permissions.all()

    def django_get_group_permissions(self, user_obj):
        user_groups_field = get_user_model()._meta.get_field('groups')
        user_groups_query = 'group__%s' % user_groups_field.related_query_name()  # noqa

        return Permission.objects.filter(**{user_groups_query: user_obj})

    def django_get_all_permissions(self, user_obj):
        if not hasattr(user_obj, '_perm_cache'):
            user_obj._perm_cache = {
                *self.django_get_permissions(user_obj, 'user'),
                *self.django_get_permissions(user_obj, 'group'),
            }

        return user_obj._perm_cache

    def django_get_permissions(self, user_obj, from_name):

        perm_cache_name = '_%s_perm_cache' % from_name
        if not hasattr(user_obj, perm_cache_name):
            if user_obj.is_superuser:
                perms = Permission.objects.all()
            else:
                perms = getattr(
                    self, 'django_get_%s_permissions' % from_name
                )(user_obj)
            perms = perms.values_list(
                'content_type__app_label', 'codename'
            ).order_by()
            setattr(
                user_obj,
                perm_cache_name,
                {"%s.%s" % (ct, name) for ct, name in perms}
            )
        return getattr(user_obj, perm_cache_name)

    def _django_has_perm(self, user_obj, perm):
        """
        Gets user permissions from associated role.
        """
        if not user_obj.is_active or user_obj.is_anonymous:
            return set()

        if user_obj.is_superuser:
            perms = Permission.objects.all()
        else:
            # this is the only major difference from django's
            # own implementation: it gets permissions from associated role
            if not user_obj.role:
                # if non superuser does not have associated a role
                # he/she basically is has no permissios at all!
                return set()
            perms = user_obj.role.permissions.all()

        # By returning permissions this way - will be able to
        # reuse ``user_obj.has_perm(...)`` method.
        _perms = perms.values_list(
            'content_type__app_label',
            'codename'
        ).order_by()

        if hasattr(user_obj, '_user_cache_perm'):
            __perms = getattr(user_obj, '_user_cache_perm')
        else:
            setattr(
                user_obj,
                '_user_cache_perm',
                {"%s.%s" % (ct, name) for ct, name in _perms}
            )
            __perms = getattr(user_obj, '_user_cache_perm')

        return perm in __perms

    def get_perms_dict(self, user_obj, obj_list, perms):
        """
        Returns a dictionary. Each key of the dictionary
        # is the id of the node. Value of the key is permissions
        # dictionary

        Input:

        user_obj = user instance core.models.User
        obj_list = a list of instances of type core.models.node.BaseTreeNode
        perms = a list of permissions e.g. ['read', 'write', delete]

        Output:

            a dictionary with given perms as keys. Example:

            ret = user.get_perms_dict([n1, n2], ['read', 'write', 'delete'])
            ret = {
                n1.id: {
                    'read': True,
                    'delete': False,
                    'write': False
                },
                n2.id: {
                    'read': False,
                    'delete': False,
                    'write': False
                }
            }

        Another example return value:

            {
              '23': {"read": True, "delete": False },
              '24': {"read": True, "delete": False },
              '25': {"read": False, "delete": False }
            }
        """
        ret = {}
        deny_perms = self._get_all_deny_permissions(
            user_obj,
            obj_list
        )

        allow_perms = self._get_all_allow_permissions(
            user_obj,
            obj_list
        )

        for obj_id in allow_perms.keys():
            ret[obj_id] = {}
            for perm in allow_perms[obj_id]:
                ret[obj_id][perm] = True

        for obj_id in deny_perms.keys():
            for perm in deny_perms[obj_id]:
                # 'deny' permissions overwrite 'allow'
                ret[obj_id][perm] = False

        return ret

    def has_perm(self, user_obj, perm, obj=None):
        """
        Main function. However it is optional in django auth backend.
        verifies if user has permissions to access a given file or folder
        via papermerge.core.models.Access models.
        """
        # Access.DENY has priority over Access.ALLOW
        if not obj:
            return self._django_has_perm(user_obj, perm)

        deny_perms = self._get_all_deny_permissions(user_obj, obj)
        allow_perms = self._get_all_allow_permissions(user_obj, obj)

        if perm in deny_perms:
            return False

        return perm in allow_perms

    def _get_group_permissions(self, user_obj, obj_or_list, access_type):
        if not isinstance(obj_or_list, models.Model):
            if len(obj_or_list) == 0:
                return set()
        else:
            if not obj_or_list:
                return {}

        if not isinstance(obj_or_list, models.Model):
            ret = {}
            obj_ids = [obj.id for obj in obj_or_list]
            all_access_items = Access.objects.prefetch_related(
                'permissions'
            ).filter(
                access_type=access_type
            ).filter(
                node_id__in=obj_ids
            )
            for access in all_access_items:
                if not access.group:
                    continue

                if user_obj.groups.filter(name=access.group.name).exists():
                    ret[access.node_id] = {
                        perm.codename for perm in access.permissions.all()
                    }

            return ret

        # else -> case when obj_or_list is a single object

        for access in obj_or_list.access_set.filter(access_type=access_type):
            if not access.group:
                continue

            if user_obj.groups.filter(name=access.group.name).exists():
                return {perm.codename for perm in access.permissions.all()}

        return set()

    def _get_user_permissions(self, user_obj, obj_or_list, access_type):
        if not isinstance(obj_or_list, models.Model):
            if len(obj_or_list) == 0:
                return set()
        else:
            if not obj_or_list:
                return {}

        if not isinstance(obj_or_list, models.Model):
            ret = {}
            obj_ids = [obj.id for obj in obj_or_list]
            all_access_items = Access.objects.prefetch_related(
                'permissions',
                'user'
            ).filter(
                access_type=access_type
            ).filter(
                node_id__in=obj_ids
            )

            for access in all_access_items:
                if access.user == user_obj:
                    ret[access.node_id] = {
                        perm.codename for perm in access.permissions.all()
                    }

            return ret

        # else -> case when obj_or_list is a single object

        for access in obj_or_list.access_set.filter(access_type=access_type):
            if access.user == user_obj:
                return {perm.codename for perm in access.permissions.all()}

        return set()

    def _get_all_allow_permissions(self, user_obj, obj_or_list):
        """
        Returns a set of permissions (python set())
        """
        all_user_perms = self._get_user_permissions(
            user_obj, obj_or_list, Access.ALLOW
        )
        all_group_perms = self._get_group_permissions(
            user_obj, obj_or_list, Access.ALLOW
        )

        ret = {}

        if not isinstance(obj_or_list, models.Model):
            for obj in obj_or_list:
                ret[obj.id] = {
                    *all_user_perms.get(obj.id, set()),
                    *all_group_perms.get(obj.id, set()),
                }
            return ret

        return {
            *all_user_perms, *all_group_perms
        }

    def _get_all_deny_permissions(self, user_obj, obj_or_list):
        """
        Returns a set of permissions (python set())
        """
        all_user_perms = self._get_user_permissions(
            user_obj,
            obj_or_list,
            Access.DENY
        )
        all_group_perms = self._get_group_permissions(
            user_obj,
            obj_or_list,
            Access.DENY
        )

        ret = {}

        if isinstance(obj_or_list, models.Model):
            return {
                *all_user_perms,
                *all_group_perms,
            }

        for obj in obj_or_list:
            ret[obj.id] = {
                *all_user_perms.get(obj.id, set()),
                *all_group_perms.get(obj.id, set()),
            }

        return ret



from django.dispatch import Signal

# signals sent by the worker
# have sender set to ``papermerge.core.signals_deginitions.WORKER``
WORKER = "worker"

automates_matching = Signal(
    providing_args=[
        "user_id",
        "level",
        "message",
        "document_id",
        "page_num",
        # text against which matching is performed
        # i.e. extracted text of the page
        "text",
        "namespace"
    ]
)

# sent by the worker
# after ocr of the page is complete i.e.
# when both .txt file and .hocr files are available
page_ocr = Signal(
    providing_args=[
        "user_id",
        "level",
        "message",
        "document_id",
        "page_num",
        "namespace",
        # status is a string: started, complete
        "status"
    ]
)

# sent by the worker before starting OCR
pre_page_ocr = Signal(
    providing_args=[
        "user_id",
        "document_id",
        "file_name",
        "page_num",
        "namespace",
    ]
)

# sent by the worker after .txt file was extracted
post_page_txt = Signal(
    providing_args=[
        "user_id",
        "document_id",
        "file_name",
        "page_num",
        "namespace",
        "text"
    ]
)

# sent by the worker after .hocr file was extracted
post_page_hocr = Signal(
    providing_args=[
        "user_id",
        "document_id",
        "fle_name",
        "page_num",
        "lang",
        "namespace",
        "step",
        "hocr"
    ]
)

# Sent by core.views.documents.create_folder
# Sent AFTER one single folder was created
folder_created = Signal(
    providing_args=[
        "user_id",
        "level",
        "message",
        "folder_id"
    ]
)

# Sent by core.views.nodes.nodes_view
# Sent AFTER one of more nodes were batch deleted
nodes_deleted = Signal(
    providing_args=[
        "user_id",
        "level",
        "message",
        # node tags is a list of
        # '<a href="{node.id}">{node.title}</a>'
        "node_tags"
    ]
)



from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views import generic
from django.urls import reverse_lazy


from papermerge.contrib.admin.models import LogEntry
from papermerge.contrib.admin.forms import LogEntryForm
from papermerge.contrib.admin.views import mixins as mix


class LogsView(
    LoginRequiredMixin,
    mix.CommonListMixin
):
    model = LogEntry
    form_class = LogEntryForm
    action_url = success_url = reverse_lazy('admin:logs')
    new_object_url = 'NA'


class LogsListView(
    LogsView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):

    title = _("Logs")

    table_header_row = [
        _('Time'),
        _('Message'),
        _('User'),
        _('Level')
    ]

    def get_queryset(self):
        qs = super().get_queryset()

        if self.request.user.is_superuser:
            # superuser sees all logs
            return qs.all()

        logs = qs.filter(user=self.request.user)

        return logs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['table_header_row'] = self.table_header_row
        return context


class LogUpdateView(LogsView, generic.UpdateView):

    title = _("Log Entry")

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = self.title

        return context



import logging
import time

from django.conf import settings
from django.core.management.base import BaseCommand
from papermerge.core.importers.local import import_documents

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """Import documents from importer directory.
    Importer directory is defined by PAPERMERGE_IMPORTER_DIR configuration
    setting.
"""

    def add_arguments(self, parser):
        parser.add_argument(
            "directory",
            default=settings.PAPERMERGE_IMPORTER_DIR,
            nargs="?",
            help="The importer directory."
        )
        parser.add_argument(
            "--loop-time",
            "-t",
            default=settings.PAPERMERGE_IMPORTER_LOOP_TIME,
            type=int,
            help="Wait time between each loop (in seconds)."
        )

    def main_loop(self, directory, loop_time):
        while True:
            start_time = time.time()
            import_documents(directory)
            # Sleep until the start of the next loop step
            time.sleep(
                max(0, start_time + loop_time - time.time())
            )

    def handle(self, *args, **options):
        directory = options.get('directory')
        loop_time = options.get('loop_time')

        logger.info(f"Watching for files in {directory}...")

        try:
            self.main_loop(directory, loop_time)
        except KeyboardInterrupt:
            logger.info("Exiting")



import logging
from django.core.management.base import BaseCommand
from papermerge.core.models import Document

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """Read .txt files and insert its content
    into associated DB column.

    It will update text field of all associated pages first
    (from .txt files) and then concatinate all text field into
    doc.text field.
"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--document-id',
            '-d',
            help="Limit update only for specified document_id"
        )

    def update_field(
        self,
        document_id=False
    ):
        docs = []
        ocred_count = 0

        if document_id:
            docs = Document.objects.filter(
                basetreenode_ptr_id=document_id
            )
        else:
            docs = Document.objects.all()

        for doc in docs:
            if doc.update_text_field():
                ocred_count += 1

    def onprem_handle(self):
        self.update_field(tenant_name=None)

    def handle(self, *args, **options):
        document_id = options.get(
            'document_id',
            False
        )
        self.update_field(
            document_id=document_id
        )



import os
import logging
import time

from django.conf import settings
from papermerge.core.storage import default_storage
from papermerge.core import signal_definitions as signals

from mglib import mime
from mglib.pdfinfo import get_pagecount
from mglib.path import (
    DocumentPath,
    PagePath,
)
from mglib.step import (Step, Steps)
from mglib.shortcuts import (
    extract_img,
    resize_img,
    extract_hocr,
    extract_txt,
)
from mglib.tiff import convert_tiff2pdf

logger = logging.getLogger(__name__)

STARTED = "started"
COMPLETE = "complete"


def notify_hocr_ready(page_path, **kwargs):
    """
    Notifies interested parties that .hocr file is available.

    Notifies via django signals. Among others will send
    hocr content itself. Input arguments:

    ``page_path``: mglib.PagePath instance of current page
    Following keys are expected to be availble in kwargs dictinary:

        * ``user_id``
        * ``document_id``
        * ``file_name``
        * ``page_num``
        * ``namespace``
        * ``version``
        * ``step``

    Always returns None.

    Sent signals: ``post_page_hocr``.

    Following arguments are passed to the signal:
        * ``sender`` = from papermerge.core.signal_definitions.WORKER
        * ``user_id``
        * ``document_id``
        * ``file_name``
        * ``page_num``
        * ``lang``
        * ``version``
        * ``namespace`` = may be empty. Used to distinguish among
            different tenants in multi-tenant deployments.
        * ``step`` = integer number corresponding to step
            learn more about steps in ``mglib.step.Step``
        * ``hocr`` = extracted hocr data (text format)
    """

    user_id = kwargs.get('user_id', None)
    document_id = kwargs.get('document_id', None)
    file_name = kwargs.get('file_name', None)
    page_num = kwargs.get('page_num', 1)
    version = kwargs.get('version', 0)
    namespace = kwargs.get('namespace', None)
    step = kwargs.get('step', 1)

    if page_path:
        abs_path_hocr = default_storage.abspath(page_path.hocr_url())

        if os.path.exists(abs_path_hocr):
            with open(abs_path_hocr) as f:
                hocr = f.read()

                signals.post_page_hocr.send(
                    sender=signals.WORKER,
                    user_id=user_id,
                    document_id=document_id,
                    file_name=file_name,
                    page_num=page_num,
                    step=step,
                    namespace=namespace,
                    version=version,
                    hocr=hocr
                )
        else:
            logger.warning(
                f"Page hocr/step={step} path {abs_path_hocr} does not exist."
            )
    else:
        logger.warning(
            f"hOCR/step={step} method returned empty page path."
        )


def notify_txt_ready(page_path, **kwargs):
    """
    Notifies interested parties that .txt file is available.

    Notifies via django signals. Among others will send
    .txt content itself. Input arguments:

    ``page_path``: mglib.PagePath instance of current page
    Following keys are expected to be availble in kwargs dictinary:

        * ``user_id``
        * ``document_id``
        * ``file_name``
        * ``page_num``
        * ``version``
        * ``namespace``

    Always returns None.

    Sent signals: ``post_page_txt``.

    Following arguments are passed to the signal:
        * ``sender`` = from papermerge.core.signal_definitions.WORKER
        * ``user_id``
        * ``document_id``
        * ``file_name``
        * ``page_num``
        * ``version``
        * ``lang``
        * ``namespace`` = may be empty. Used to distinguish among
            different tenants in multi-tenant deployments.
        * ``txt`` = extracted .txt data (text format)
    """

    user_id = kwargs.get('user_id', None)
    document_id = kwargs.get('document_id', None)
    page_num = kwargs.get('page_num', 1)
    file_name = kwargs.get('file_name', None)
    version = kwargs.get('version', 0)
    namespace = kwargs.get('namespace', None)

    logger.debug("notify_txt_ready")

    if page_path:
        abs_path_txt = default_storage.abspath(page_path.txt_url())

        logger.debug(f"notify_txt_ready for {abs_path_txt}")

        if os.path.exists(abs_path_txt):
            with open(abs_path_txt) as f:
                text = f.read()

                logger.debug(
                    f"Sending post_page_txt signal"
                    f" namespace={namespace} "
                    f" user_id={user_id}"
                    f" document_id={document_id}"
                    f" page_num={page_num}"
                    f" text={text}"
                )
                signals.post_page_txt.send(
                    sender=signals.WORKER,
                    user_id=user_id,
                    document_id=document_id,
                    file_name=file_name,
                    page_num=page_num,
                    version=version,
                    namespace=namespace,
                    text=text
                )
        else:
            logger.warning(
                f"Page txt path {abs_path_txt} does not exist. "
                f"Page indexing was skipped."
            )
    else:
        logger.warning(
            "OCR method returned empty page path. "
            "Page indexing was skipped."
        )


def notify_pre_page_ocr(page_path, **kwargs):
    user_id = kwargs.get('user_id', None)
    document_id = kwargs.get('document_id', None)
    file_name = kwargs.get('file_name', None)
    page_num = kwargs.get('page_num', 1)
    version = kwargs.get('version', 0)
    namespace = kwargs.get('namespace', None)

    signals.pre_page_ocr.send(
        sender=signals.WORKER,
        user_id=user_id,
        document_id=document_id,
        file_name=file_name,
        page_num=page_num,
        version=version,
        namespace=namespace,
    )


def ocr_page_pdf(
    doc_path,
    page_num,
    lang,
    **kwargs
):
    """
    doc_path is an mglib.path.DocumentPath instance

    On success returns ``mglib.path.PagePath`` instance.
    """
    logger.debug("OCR PDF document")

    file_name = kwargs.pop('file_name', None)

    if not file_name:
        file_name = doc_path.file_name

    page_count = get_pagecount(
        default_storage.abspath(doc_path.url())
    )

    if page_num <= page_count:
        # first quickly generate preview images
        page_path = PagePath(
            document_path=doc_path,
            page_num=page_num,
            step=Step(1),
            page_count=page_count
        )
        for step in Steps():
            page_path.step = step
            extract_img(
                page_path,
                media_root=settings.MEDIA_ROOT
            )

    notify_pre_page_ocr(
        page_path,
        page_num=page_num,
        lang=lang,
        file_name=doc_path.file_name,
        **kwargs
    )

    if page_num <= page_count:
        page_path = PagePath(
            document_path=doc_path,
            page_num=page_num,
            step=Step(1),
            page_count=page_count
        )
        extract_txt(
            page_path,
            lang=lang,
            media_root=settings.MEDIA_ROOT
        )
        notify_txt_ready(
            page_path,
            page_num=page_num,
            lang=lang,
            file_name=file_name,
            **kwargs
        )

        for step in Steps():
            page_path.step = step
            if not step.is_thumbnail:
                extract_hocr(
                    page_path,
                    lang=lang,
                    media_root=settings.MEDIA_ROOT
                )
                notify_hocr_ready(
                    page_path,
                    page_num=page_num,
                    lang=lang,
                    # step as integer number
                    step=step.current,
                    file_name=file_name,
                    **kwargs
                )

    return page_path


def ocr_page_image(
    doc_path,
    page_num,
    lang,
    **kwargs
):
    """
    image = jpg, jpeg, png

    On success returns ``mglib.path.PagePath`` instance.
    """
    logger.debug("OCR image (jpeg, jpg, png) document")

    page_path = PagePath(
        document_path=doc_path,
        page_num=page_num,
        step=Step(1),
        # jpeg, jpg, png are 1 page documents
        page_count=1
    )
    notify_pre_page_ocr(
        page_path,
        page_num=page_num,
        lang=lang,
        file_name=doc_path.file_name,
        **kwargs
    )
    # resize and eventually convert (png -> jpg)
    resize_img(
        page_path,
        media_root=settings.MEDIA_ROOT
    )
    extract_txt(
        page_path,
        lang=lang,
        media_root=settings.MEDIA_ROOT
    )
    notify_txt_ready(
        page_path,
        page_num=page_num,
        lang=lang,
        file_name=doc_path.file_name,
        **kwargs
    )

    # First quickly generate preview images
    for step in Steps():
        page_path.step = step
        resize_img(
            page_path,
            media_root=settings.MEDIA_ROOT
        )
    # reset page's step
    page_path.step = Step(1)
    # Now OCR each image
    for step in Steps():
        if not step.is_thumbnail:
            extract_hocr(
                page_path,
                lang=lang,
                media_root=settings.MEDIA_ROOT
            )
            notify_hocr_ready(
                page_path,
                page_num=page_num,
                lang=lang,
                # step as integer number
                step=step.current,
                file_name=doc_path.file_name,
                **kwargs
            )

    return page_path


def ocr_page(
    user_id,
    document_id,
    file_name,
    page_num,
    lang,
    version,
    namespace=None,
):
    logger.debug(
        f" ocr_page user_id={user_id} doc_id={document_id}"
        f" page_num={page_num}"
    )
    t1 = time.time()
    lang = lang.lower()
    doc_path = DocumentPath(
        user_id=user_id,
        document_id=document_id,
        file_name=file_name,
        version=version
    )

    if not default_storage.exists(doc_path.url()):
        # In case of distibuted deployment, document uploaded
        # by webapp is not directly available to the worker (which runs on
        # separate computer). Thus, if document is not locally available,
        # worker will download the document from whatever remote location.
        default_storage.download(
            doc_path_url=doc_path.url(),
            namespace=namespace
        )

    mime_type = mime.Mime(
        default_storage.abspath(doc_path.url())
    )
    logger.debug(f"Mime Type = {mime_type}")

    page_type = ''

    if mime_type.is_pdf():
        ocr_page_pdf(
            doc_path=doc_path,
            page_num=page_num,
            lang=lang,
            user_id=user_id,
            version=version,
            document_id=document_id,
            namespace=namespace
        )
        page_type = 'pdf'
    elif mime_type.is_image():  # jpeg, jpeg or png
        ocr_page_image(
            doc_path=doc_path,
            page_num=page_num,
            lang=lang,
            user_id=user_id,
            document_id=document_id,
            namespace=namespace,
            version=version
        )
    elif mime_type.is_tiff():
        # new filename is a pdf file
        logger.debug("TIFF type detected")
        new_filename = convert_tiff2pdf(
            doc_url=default_storage.abspath(doc_path.url())
        )
        # now .pdf
        orig_file_name = doc_path.file_name
        doc_path.file_name = new_filename
        # and continue as usual
        ocr_page_pdf(
            doc_path=doc_path,
            page_num=page_num,
            lang=lang,
            user_id=user_id,
            document_id=document_id,
            # Pass original file_name i.e. tiff file name as well.
            file_name=orig_file_name,
            namespace=namespace,
            version=version
        )
    else:
        logger.error(
            f" user_id={user_id}"
            f" doc_id={document_id}"
            f" page_num={page_num} error=Unkown file type"
        )
        return True

    t2 = time.time()
    logger.debug(
        f" user_id={user_id} doc_id={document_id}"
        f" page_num={page_num} page_type={page_type}"
        f" total_exec_time={t2-t1:.2f}"
    )

    return True



# Generated by Django 3.0.6 on 2020-06-02 15:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0010_remove_kvstore_human_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='kvstorenode',
            name='node',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.BaseTreeNode'),
        ),
    ]



import lxml
import lxml.html
import re
import logging


logger = logging.getLogger(__name__)


def extract_size(title):
    box_re = re.compile(
        r'.*; bbox (?P<x1>\d+) (?P<y1>\d+) (?P<x2>\d+) (?P<y2>\d+);.*'
    )
    width = None
    height = None
    matched_obj = re.match(box_re, title)
    if matched_obj:
        width = int(matched_obj['x2'])
        height = int(matched_obj['y2'])

    return width, height


class OcrxWord:
    def __init__(self, el_class, el_id, title, text=''):
        """
        span_elem is item from array returnd by lxml.html's xpath
        method
        """
        self.el_class = el_class
        self.el_id = el_id
        self.title = title
        self.text = text
        self.x1 = 0
        self.x2 = 0
        self.y1 = 0
        self.y2 = 0
        self.wconf = 0
        self.build_bbox_attrs(self.title)

    def build_bbox_attrs(self, title):
        """
        build x1, x2, y1, y2 and wconf attribues from title string.

        Example of title strings:

        'bbox 102 448 120 457; x_wconf 38'
        """
        box_re = re.compile(
            r'bbox (?P<x1>\d+) (?P<y1>\d+)'
            r' (?P<x2>\d+) (?P<y2>\d+); x_wconf (?P<wconf>\d+)'
        )
        matched_obj = re.match(box_re, title)
        if matched_obj:
            self.x1 = int(matched_obj['x1'])
            self.y1 = int(matched_obj['y1'])
            self.x2 = int(matched_obj['x2'])
            self.y2 = int(matched_obj['y2'])
            self.wconf = int(matched_obj['wconf'])
        else:
            logger.info(
                f"Word title mismatch.title={self.title}, el_id={self.el_id}."
            )

    def to_hash(self):
        """
        returns hash structure from OcrxWord attributes.
        """
        return {
            'x1': self.x1,
            'y1': self.y1,
            'x2': self.x2,
            'y2': self.y2,
            'wconf': self.wconf,
            'id': self.el_id,
            'title': self.title,
            'text': self.text,
            'class': self.el_class
        }


class Hocr:
    """Manages ocrx words from hocr file.
    """

    def __init__(self, hocr_file_path, min_wconf=30):
        self.hocr_file_path = hocr_file_path
        self.ocrx_words = []
        self.min_wconf = min_wconf
        self._width = 0
        self._height = 0
        try:
            self.extract()
        except lxml.etree.ParserError:
            logger.warning(
                f"Hocr file {hocr_file_path}"
                " is either empty of not of HOCR format"
            )

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def extract(self):
        html = None

        with open(self.hocr_file_path, "rb") as f:
            text = f.read()
            html = lxml.html.fromstring(text)

        page = html.xpath("//div[@class='ocr_page']")
        if len(page) > 0:
            self._width, self._height = extract_size(
                page[0].attrib['title']
            )

        for span in html.xpath("//span[@class='ocrx_word']"):
            elem = OcrxWord(
                el_class=span.attrib['class'],
                el_id=span.attrib['id'],
                title=span.attrib['title'],
                text=span.text
            )
            self.ocrx_words.append(elem)

        return self.ocrx_words

    def good_json_words(self):
        """Return only:
            * ocrx words with non empty text
            * ocrx words with w_conf > min_w_conf
        """
        meta = self._filter_words()
        return meta['good_words']

    def _filter_words(self):
        meta = {}
        meta['count_all'] = len(self.ocrx_words)
        meta['bad_words'] = []
        meta['good_words'] = []
        meta['count_non_empty'] = 0
        meta['count_low_wconf'] = 0
        meta['width'] = self.width
        meta['height'] = self.height
        meta['min_wconf'] = self.min_wconf

        for word in self.ocrx_words:
            bad_word = False
            if word.text and len(word.text) == 0:
                bad_word = True
                meta['bad_words'].append(word.to_hash())
                meta['count_non_empty'] += 1

            if word.wconf < self.min_wconf:
                bad_word = True
                meta['bad_words'].append(word.to_hash())
                meta['count_low_wconf'] += 1

            if not bad_word:
                meta['good_words'].append(word.to_hash())

        meta['count_bad'] = len(meta['bad_words'])
        meta['count_good'] = len(meta['good_words'])

        return meta

    def get_meta(self):
        meta = self._filter_words()

        return {
            'width': meta['width'],
            'height': meta['height'],
            'count_all': meta['count_all'],
            'count_bad': meta['count_bad'],
            'count_good': meta['count_good'],
            'count_non_empty': meta['count_non_empty'],
            'count_low_wconf': meta['count_low_wconf'],
            'bad_words': meta['bad_words'],
            'min_wconf': meta['min_wconf']
        }



# Generated by Django 3.0.8 on 2020-09-09 08:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0021_tag_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='tag',
            name='pinned',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='tag',
            name='bg_color',
            field=models.CharField(default='#c41fff', max_length=7, verbose_name='Background Color'),
        ),
        migrations.AlterField(
            model_name='tag',
            name='description',
            field=models.TextField(blank=True, max_length=1024, null=True, verbose_name='Description (optional)'),
        ),
        migrations.AlterField(
            model_name='tag',
            name='fg_color',
            field=models.CharField(default='#FFFFFF', max_length=7, verbose_name='Foreground Color'),
        ),
    ]



import json
import logging

from django.contrib.auth.decorators import login_required
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseForbidden
)

from papermerge.core.models import (
    BaseTreeNode,
    Page,
    Access
)

from papermerge.core.models.kvstore import (get_currency_formats,
                                            get_date_formats, get_kv_types,
                                            get_numeric_formats)

from .utils import sanitize_kvstore_list

logger = logging.getLogger(__name__)


@login_required
def metadata(request, model, id):
    """
    model can be either node or page. Respectively
    id will be the 'id' of either node or page.
    E.g.
    POST /metadata/page/55 # will update metadata for page id=55
    POST /metadata/node/40 # will update metadata for node id=40
    """
    node = None
    if model == 'node':
        _Klass = BaseTreeNode
        # if queried node is a document, then instead
        # of document's metadata - metadata of first page
        # of respective document will be returned.
        node = BaseTreeNode.objects.get(id=id)
        if node.is_document():
            # will return first page
            # i.e. page with lowerest page.number
            # attribute
            _Klass = Page
            id = node.pages.first().id
    else:
        _Klass = Page

    try:
        item = _Klass.objects.get(id=id)
    except _Klass.DoesNotExist:
        raise Http404("Node does not exists")

    # allow access to metadata only if user has read permissions
    # on the document.
    document_or_node = node
    if model == 'page':
        document_or_node = item.document

    if not request.user.has_perm(
        Access.PERM_READ,
        document_or_node
    ):
        return HttpResponseForbidden()

    kvstore = []

    if request.method == 'GET':
        for kv in item.kv.all():
            kvstore.append(kv.to_dict())
    else:
        # Note that for a document metadata can be updated in two ways:
        #
        #   1) POST /metadata/node/:id - request issued from document browser
        #        :id is the ``id`` of respective node
        #   2) POST /metadata/page/:id i.e. - request issued from document
        #       viewer. :id in this case is ``id`` of first page of the
        #       document.
        #
        # For case 1) the key/value will be applied on the document and
        # propagated to the first page while for 2) the update is applied to
        # the page and not propagated as page models are leaf elements.
        #
        # This slight inconsistency works because document metadata is not used
        # at all. Instead what user thinks is document's metadata is actually
        # metadata of the first page of the document.
        if isinstance(item, BaseTreeNode):
            node = item
        else:
            node = item.document

        if request.user.has_perm(Access.PERM_WRITE, node):
            kv_data = json.loads(request.body)
            if 'kvstore' in kv_data:
                if isinstance(kv_data['kvstore'], list):
                    item.kv.update(
                        sanitize_kvstore_list(kv_data['kvstore'])
                    )
        else:
            return HttpResponseForbidden()

    return HttpResponse(
        json.dumps(
            {
                'kvstore': kvstore,
                'currency_formats': get_currency_formats(),
                'date_formats': get_date_formats(),
                'numeric_formats': get_numeric_formats(),
                'kv_types': get_kv_types()

            }
        ),
        content_type="application/json"
    )



import os
import logging

from celery import Celery
from celery.apps.worker import Worker as CeleryWorker
from django.core.management.base import BaseCommand
from django.conf import settings
from papermerge.core.models import Document, BaseTreeNode
from papermerge.core.importers.imap import import_attachment
from papermerge.core.importers.local import import_documents

logger = logging.getLogger(__name__)
celery_app = Celery('papermerge')


@celery_app.task
def rebuild_the_tree():
    # https://github.com/django-mptt/django-mptt/issues/568
    BaseTreeNode.objects.rebuild()


@celery_app.task
def txt2db():
    """
    Move OCRed text from txt files into database
    """
    logger.debug("Celery beat: txt2db")

    for doc in Document.objects.all():
        doc.update_text_field()


@celery_app.task
def import_from_email():
    """
    Import attachments from specified email account.
    """
    # if no email import defined, just skip the whole
    # thing.
    if not settings.PAPERMERGE_IMPORT_MAIL_USER:
        return

    logger.debug("Celery beat: import_from_email")

    imap_server = settings.PAPERMERGE_IMPORT_MAIL_HOST
    username = settings.PAPERMERGE_IMPORT_MAIL_USER
    password = settings.PAPERMERGE_IMPORT_MAIL_PASS
    delete = settings.PAPERMERGE_IMPORT_MAIL_DELETE
    by_user = settings.PAPERMERGE_IMPORT_MAIL_BY_USER
    by_secret = settings.PAPERMERGE_IMPORT_MAIL_BY_SECRET
    inbox_name = settings.PAPERMERGE_IMPORT_MAIL_INBOX

    import_attachment(
        imap_server=imap_server,
        username=username,
        password=password,
        delete=delete,
        inbox_name=inbox_name,
        by_user=by_user,
        by_secret=by_secret
    )


@celery_app.task
def import_from_local_folder():
    """
    Import documents from defined local folder
    """
    logger.debug("Celery beat: import_from_local_folder")
    import_documents(settings.PAPERMERGE_IMPORTER_DIR)


def _include_txt2db_task(celery_app_instance, schedule):
    # Calls every :schedule: seconds txt2db

    celery_app_instance.add_periodic_task(
        schedule,  # call this task once in :schedule: seconds
        txt2db.s(),
        name='txt2db'
    )


def _include_rebuid_tree_task(celery_app_instance, schedule):
    # once every :schedule: seconds rebuild the whole tree

    celery_app_instance.add_periodic_task(
        schedule,  # call this task once in :schedule: seconds
        rebuild_the_tree.s(),
        name='rebuild_the_tree'
    )


def _include_local_dir_task(celery_app_instance):
    imp_dir_exists = None
    imp_dir = settings.PAPERMERGE_IMPORTER_DIR
    if settings.PAPERMERGE_IMPORTER_DIR:
        imp_dir_exists = os.path.exists(settings.PAPERMERGE_IMPORTER_DIR)

    if imp_dir and imp_dir_exists:
        celery_app_instance.add_periodic_task(
            30.0,
            import_from_local_folder.s(),
            name='import_from_local_folder'
        )
    else:
        reason_msg = ""

        if not imp_dir:
            reason_msg += "1. IMPORTER_DIR not configured\n"
        if not imp_dir_exists:
            reason_msg += "2. importer dir does not exist\n"

        logger.warning(
            "Importer from local folder task not started."
            "Reason:\n" + reason_msg
        )


def _include_imap_import_task(celery_app_instance):
    mail_user = settings.PAPERMERGE_IMPORT_MAIL_USER
    mail_host = settings.PAPERMERGE_IMPORT_MAIL_HOST

    if mail_user and mail_host:
        celery_app_instance.add_periodic_task(
            30.0, import_from_email.s(), name='import_from_email'
        )
    else:
        reason_msg = ""
        if not mail_user:
            reason_msg += "PAPERMERGE_IMPORT_MAIL_USER not defined\n"
        if not mail_host:
            reason_msg += "PAPERMERGE_IMPORT_MAIL_HOST not defined\n"

        logger.warning(
            "Importer from imap account not started."
            "Reason:\n" + reason_msg
        )


def setup_periodic_tasks(celery_app_instance, **options):
    """
    ``options`` will have following keys:

        * skip_imap :boolean:
        * skip_local :boolean:
        * skip_txt2db :boolean:
        * skip_rebuid :boolean:

    Each of those keys decides weathe to start respective period task.
    Depending on setup, user might want to skip one, two or maybe all four
    period tasks.
    """

    # if user provides --skip-imap flag i.e.
    #
    #   $ ./manage.py worker --skip-imap
    #
    # then
    #   1. options.get('skip_imap') is True
    #   2. not options.get('skip_imap') is False
    #   3. start_imap_import = False
    start_imap_import = not options.get("skip_imap", False)
    start_local_import = not options.get("skip_local", False)
    start_txt2db = not options.get("skip_txt2db", False)
    start_rebuild = not options.get("skip_rebuild", False)
    rebuild_schedule = options.get("rebuild_schedule")
    txt2db_schedule = options.get("txt2db_schedule")

    if start_txt2db:
        _include_txt2db_task(
            celery_app_instance,
            schedule=txt2db_schedule
        )

    if start_rebuild:
        _include_rebuid_tree_task(
            celery_app_instance,
            schedule=rebuild_schedule
        )

    if start_local_import:
        _include_local_dir_task(celery_app_instance)

    if start_imap_import:
        _include_imap_import_task(celery_app_instance)


class Command(BaseCommand):

    help = "Built-in Papermerge worker"

    def add_arguments(self, parser):
        parser.add_argument(
            '--pidfile',
            type=str,
            help='Optional file used to store the process pid.\n'
            'The program won’t start if this file already '
            'exists and the pid is still alive.'
        )
        parser.add_argument(
            '--skip-imap',
            action="store_true",
            help="Skip IMAP importing part. "
            "With this flag option present, worker won't "
            "start period task for fetching attachments from "
            "IMAP account."
        )
        parser.add_argument(
            '--skip-local',
            action="store_true",
            help="Skip local directory importing part. "
            "With this flag option present, worker won't "
            "start period task for ingesting documents from "
            "local folder."
        )
        parser.add_argument(
            '--skip-rebuild',
            action="store_true",
            help="Do not start 'rebuild_the_tree' periodic task."
        )
        parser.add_argument(
            '--rebuild-schedule',
            type=int,
            default=300,
            help="Schedule (in seconds) for 'rebuild_the_tree' periodic task."
            " Default value is 300 seconds."
        )
        parser.add_argument(
            '--skip-txt2db',
            action="store_true",
            help="Do not start 'txt2db' periodic task."
        )
        parser.add_argument(
            '--txt2db-schedule',
            type=int,
            default=64,
            help="Schedule (in seconds) for 'txt2db' periodic task."
            " Default value is 64 seconds."
        )

    def handle(self, *args, **options):
        celery_app.config_from_object(
            'django.conf:settings', namespace='CELERY'
        )
        # Load task modules from all registered Django app configs.
        celery_app.autodiscover_tasks()

        celery_worker = CeleryWorker(
            hostname="localhost",
            app=celery_app,
            beat=True,
            quiet=True,
            concurrency=1
        )

        # Set pidfile if it the corresponding argument has been provided
        if options['pidfile']:
            celery_worker.pidfile = options['pidfile']

        setup_periodic_tasks(
            celery_app_instance=celery_app,
            **options
        )

        celery_worker.start()



import os
import time
import logging
from hashlib import blake2b

from django.conf import settings

from papermerge.core.import_pipeline import LOCAL, go_through_pipelines

logger = logging.getLogger(__name__)


def import_documents(directory, skip_ocr=False):
    files = []

    if not directory:
        raise ValueError("Import directory value is None")

    for entry in os.scandir(directory):
        if entry.is_file():
            with open(entry.path, 'rb') as file_handler:
                file_hash = blake2b()
                file_bytes = file_handler.read()
                file_hash.update(file_bytes)
            file_tuple = (entry.path, file_hash.digest())
            files.append(file_tuple)
        else:
            logger.warning(
                "Skipping %s as it is not a file",
                entry.path
            )

    if not files:
        return

    time.sleep(int(settings.PAPERMERGE_FILES_MIN_UNMODIFIED_DURATION))

    for file_path, file_hash in files:
        with open(file_path, 'rb') as file_handler:
            file_bytes = file_handler.read()
            file_hash_new = blake2b()
            file_hash_new.update(file_bytes)
            if not file_hash == file_hash_new.digest():
                continue
            # File has not been modified and can be consumed
            basename = os.path.basename(file_path)
            init_kwargs = {'payload': file_bytes, 'processor': LOCAL}
            apply_kwargs = {'user': None,
                            'name': basename,
                            'skip_ocr': skip_ocr
                            }
            doc = go_through_pipelines(init_kwargs, apply_kwargs)
            if doc is not None:
                os.remove(file_path)



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import TemplateView

from papermerge.search.backends import get_search_backend
from papermerge.core.models import (
    Page,
    Folder,
    BaseTreeNode,
    Access,
)

from papermerge.core.views.decorators import json_response
from papermerge.core import __version__ as PAPERMERGE_VERSION


class BrowseView(
    LoginRequiredMixin,
    TemplateView,
):
    template_name = "admin/index.html"


@login_required
def inbox_view(request):

    try:
        inbox = Folder.objects.get(
            title=Folder.INBOX_NAME,
            user=request.user
        )
        root_node_id = inbox.id

    except Folder.DoesNotExist:
        root_node_id = None

    return render(
        request,
        "admin/index.html",
        {
            'root_node_id': root_node_id
        }
    )


@json_response
@login_required
def version_view(request):
    return PAPERMERGE_VERSION


@login_required
def search(request):
    # 'qa' is search term from "advanced search form"
    search_term = request.GET.get('q') or request.GET.get('qa')
    tags_list = request.GET.getlist('tag')
    tags_op = request.GET.get('tags_op')
    folder_id = request.GET.get('folder', None)
    folder = None
    descendant_ids = []
    backend = get_search_backend()

    results_folders = backend.search(search_term, Folder)
    results_docs = backend.search(search_term, Page)

    if folder_id:
        try:
            folder = BaseTreeNode.objects.get(id=folder_id)
        except BaseTreeNode.DoesNotExist:
            pass

    if folder:
        descendant_ids = [node.id for node in folder.get_descendants()]

    if results_docs:
        qs_docs = BaseTreeNode.objects.filter(
            id__in=[
                p.document.basetreenode_ptr_id for p in results_docs
            ]
        )
    elif search_term:
        qs_docs = BaseTreeNode.objects.none()
    else:
        qs_docs = BaseTreeNode.objects.all()

    if descendant_ids:
        qs_docs = qs_docs.filter(id__in=descendant_ids)

    if tags_list:
        # all tags
        if tags_op == 'any':
            qs_docs = qs_docs.filter(
                tags__name__in=tags_list
            ).distinct()
        else:  # tags_op == all
            for tag in tags_list:
                qs_docs = qs_docs.filter(
                    tags__name__in=[tag]
                ).distinct()

    nodes_perms = request.user.get_perms_dict(
        qs_docs, Access.ALL_PERMS
    )

    qs_docs = [
        node for node in qs_docs.prefetch_related('tags')
        if nodes_perms[node.id].get(Access.PERM_READ, False)
    ]

    # filter out all documents for which current user
    # does not has read access

    return render(
        request,
        "admin/search.html",
        {
            'results_docs': qs_docs,
            'results_folders': results_folders.results(),
            'search_term': search_term,
            'tags_list': tags_list,
            'folder': folder,
            'tags_op': tags_op
        }
    )



import pytz

from django.utils import timezone
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.models import ContentType

from taggit.managers import TaggableManager

from papermerge.core import validators
from papermerge.core.models.access import Access
from papermerge.core.models.diff import Diff
from papermerge.core.models.tags import ColoredTag
from papermerge.core.models.kvstore import KVStoreCompNode, KVStoreNode
from polymorphic_tree.models import (
    PolymorphicMPTTModel,
    PolymorphicTreeForeignKey
)
from polymorphic_tree.managers import (
    PolymorphicMPTTModelManager,
    PolymorphicMPTTQuerySet
)

# things you can propagate from parent node to
# child node
PROPAGATE_ACCESS = 'access'
PROPAGATE_KV = 'kv'
PROPAGATE_KVCOMP = 'kvcomp'
RELATED_NAME_FMT = "%(app_label)s_%(class)s_related"
RELATED_QUERY_NAME_FMT = "%(app_label)s_%(class)ss"


class NodeManager(PolymorphicMPTTModelManager):
    pass


class NodeQuerySet(PolymorphicMPTTQuerySet):

    def delete(self, *args, **kwargs):
        for node in self:
            descendants = node.get_descendants()

            if descendants.count() > 0:
                descendants.delete(*args, **kwargs)
            # At this point all descendants were deleted.
            # Self delete :)
            try:
                node.delete(*args, **kwargs)
            except BaseTreeNode.DoesNotExist:
                # this node was deleted by earlier recursive call
                # it is ok, just skip
                pass


CustomNodeManager = NodeManager.from_queryset(NodeQuerySet)


class BaseTreeNode(PolymorphicMPTTModel):
    parent = PolymorphicTreeForeignKey(
        'self',
        models.CASCADE,
        blank=True,
        null=True,
        related_name='children',
        verbose_name=_('parent')
    )
    title = models.CharField(
        _("Title"),
        max_length=200,
        validators=[validators.safe_character_validator]
    )

    lang = models.CharField(
        _('Language'),
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )

    user = models.ForeignKey('User', models.CASCADE)

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    tags = TaggableManager(through=ColoredTag)

    # custom Manager + custom QuerySet
    objects = CustomNodeManager()

    def human_datetime(self, _datetime) -> str:
        """
        Localize and format datetime instance considering user preferences.
        """
        tz = pytz.timezone(
            self.user.preferences['localization__timezone']
        )
        fmt = self.user.preferences['localization__date_format']
        fmt += " " + self.user.preferences['localization__time_format']

        ret_datetime = timezone.localtime(_datetime, timezone=tz)

        ret = ret_datetime.strftime(fmt)
        return ret

    @property
    def human_updated_at(self) -> str:
        """
        updated_at displayed considering user's timezone, date and time prefs.

        returns string with user friendly formated datetime.
        """

        ret = self.human_datetime(self.updated_at)
        return ret

    @property
    def human_created_at(self) -> str:
        """
        created_at displayed considering user's timezone, date and time prefs.
        """
        ret = self.human_datetime(self.created_at)
        return ret

    def is_folder(self):
        folder_ct = ContentType.objects.get(
            app_label='core', model='folder'
        )
        return self.polymorphic_ctype_id == folder_ct.id

    def is_document(self):
        document_ct = ContentType.objects.get(
            app_label='core', model='document'
        )
        return document_ct.id == self.polymorphic_ctype_id

    def _get_access_diff_updated(self, new_access_list=[]):
        """
        gathers Diff with updated operation
        """
        updates = Diff(op=Diff.UPDATE)

        for current in self.access_set.all():
            for new_access in new_access_list:
                if new_access == current and current.perm_diff(new_access):
                    updates.add(new_access)

        return updates

    def _get_access_diff_deleted(self, new_access_list=[]):
        """
        gathers Diff with deleted operation
        """
        dels = Diff(op=Diff.DELETE)

        # if current access is not in the new list
        # it means current access was deleted.
        for current in self.access_set.all():
            if current not in new_access_list:
                dels.add(current)

        return dels

    def _get_access_diff_added(self, new_access_list=[]):
        """
        gathers Diff with added operation
        """
        adds = Diff(op=Diff.ADDED)

        # if current access is not in the new list
        # it means current access was deleted.
        all_current = self.access_set.all()

        # if new access is not in current access list of the
        # node - it means it will be added.
        for new_access in new_access_list:
            if new_access not in all_current:
                adds.add(new_access)

        return adds

    def get_access_diffs(self, new_access_list=[]):
        """
        * access_list is a list of Access model instances.

        Returns a list of instances of AccessDiff
        between current node access list and given access list.

        Returned value will be empty list if access_list is same
        as node's current access.
        Returned value will be a list with one entry if one or
        several access were added.
        Returned value will be a list with one entry if one or
        several access were removed.
        Returned value will be a list with two entries if one or
        several access were added and one or several entries were
        removed.
        """
        ret = []
        ret.append(
            self._get_access_diff_updated(new_access_list)
        )
        ret.append(
            self._get_access_diff_added(new_access_list)
        )
        ret.append(
            self._get_access_diff_deleted(new_access_list)
        )

        return ret

    def _apply_diff_add(self, diff):
        model = diff.first()
        if isinstance(model, Access):
            for _model in diff:
                Access.create(
                    node=self,
                    access_inherited=True,
                    access=_model
                )
        elif isinstance(model, KVStoreNode):

            array_to_apply = []

            for _model in diff:
                array_to_apply.append(
                    {
                        'kv_inherited': True,
                        'key': _model.key,
                        'kv_type': _model.kv_type,
                        'kv_format': _model.kv_format
                    }
                )

            self.kv.apply_additions(array_to_apply)

        elif isinstance(model, KVStoreCompNode):
            pass
        else:
            raise ValueError(
                f"Don't know how to replace {model} (found in {diff})"
            )

    def _apply_diff_delete(self, diff):
        ids_to_delete = []
        model = diff.first()

        if isinstance(model, Access):
            for existing_model in self.access_set.all():
                for new_model in diff:
                    if existing_model == new_model:
                        ids_to_delete.append(
                            existing_model.id
                        )
            Access.objects.filter(id__in=ids_to_delete).delete()
        elif isinstance(model, KVStoreNode):
            self.kv.apply_deletions(
                [
                    {'kv_inherited': True, 'key': _model.key}
                    for _model in diff
                ]
            )
        elif isinstance(model, KVStoreCompNode):
            pass
        else:
            raise ValueError(
                f"Don't know how to replace {model} (found in {diff})"
            )

    def _apply_diff_update(self, diff, attr_updates):
        model = diff.first()
        if isinstance(model, Access):
            for existing_model in self.access_set.all():
                for new_model in diff:
                    existing_model.update_from(new_model)
        elif isinstance(model, KVStoreNode):

            if len(attr_updates) > 0:
                updates = attr_updates
            else:
                updates = [{
                    'kv_inherited': True,
                    'key': _model.key,
                    'kv_format': _model.format,
                    'kv_type': _model.type,
                    'id': _model.id
                } for _model in diff]

            self.kv.apply_updates(updates)
        elif isinstance(model, KVStoreCompNode):
            pass
        else:
            raise ValueError(
                f"Don't know how to replace {model} (found in {diff})"
            )

    def apply_diff(self, diff, attr_updates):
        if diff.is_add():
            self._apply_diff_add(diff)
        elif diff.is_update():
            self._apply_diff_update(diff, attr_updates)
        elif diff.is_delete():
            self._apply_diff_delete(diff)
        else:
            raise ValueError("Unexpected diff {diff} type")

    def replace_access(self, diff):
        # delete exiting
        self.access_set.all().delete()

        # replace with new ones
        for access in diff:
            Access.create(
                node=self,
                access_inherited=True,
                access=access
            )

    def replace_kv(self, diff):
        pass

    def replace_kvcomp(self, diff):
        pass

    def replace_diff(self, diff):
        model = diff.first()

        if isinstance(model, Access):
            self.replace_access(diff)
        elif isinstance(model, KVStoreNode):
            # replace associated kv of current node
            # with new one specified by diff elements
            self.replace_kv(diff)
        elif isinstance(model, KVStoreCompNode):
            # replace associated kvcomp of current node
            # with new one specified by diff elements
            self.replace_kvcomp(diff)
        else:
            raise ValueError(
                f"Don't know how to replace {model} (found in {diff})"
            )

    def apply_diffs(self, diffs_list, attr_updates):
        for diff in diffs_list:
            if diff.is_update() or diff.is_add() or diff.is_delete():
                # add (new), update (existing) or delete(existing)
                self.apply_diff(diff, attr_updates)
            elif diff.is_replace():
                self.replace_diff(diff)

    def update_kv(self, key, operation):
        pass

    def propagate_changes(
        self,
        diffs_set,
        apply_to_self,
        attr_updates=[]
    ):
        """
        Recursively propagates list of diffs
        (i.e. apply changes to all children).
        diffs_set is a set of papermerge.core.models.Diff instances
        """

        if apply_to_self:
            self.apply_diffs(
                diffs_set,
                attr_updates=attr_updates
            )

        children = self.get_children()
        if children.count() > 0:
            for node in children:
                node.apply_diffs(
                    diffs_set,
                    attr_updates=attr_updates
                )
                node.propagate_changes(
                    diffs_set,
                    apply_to_self=False,
                    attr_updates=attr_updates
                )

    class Meta(PolymorphicMPTTModel.Meta):
        # please do not confuse this "Documents" verbose name
        # with real Document object, which is derived from BaseNodeTree.
        # The reason for this naming confusing is that from the point
        # of view of users, the BaseNodeTree are just a list of documents.
        verbose_name = _("Documents")
        verbose_name_plural = _("Documents")
        _icon_name = 'basetreenode'


class AbstractNode(models.Model):
    """
    Common class apps need to inherit from in order
    to extend BaseTreeNode model.
    """
    base_ptr = models.ForeignKey(
        BaseTreeNode,
        related_name=RELATED_NAME_FMT,
        related_query_name=RELATED_QUERY_NAME_FMT,
        on_delete=models.CASCADE
    )

    class Meta:
        abstract = True

    def get_title(self):
        return self.base_ptr.title



# Generated by Django 3.1.3 on 2020-12-07 10:37

from django.db import migrations, models
import django.db.models.deletion
import papermerge.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0034_auto_20201201_1116'),
    ]

    operations = [
        migrations.AlterField(
            model_name='basetreenode',
            name='title',
            field=models.CharField(max_length=200, validators=[papermerge.core.validators.safe_character_validator], verbose_name='Title'),
        ),
        migrations.AlterField(
            model_name='basetreenode',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name='basetreenode',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='document',
            name='file_name',
            field=models.CharField(default='', max_length=1024, validators=[papermerge.core.validators.safe_character_validator]),
        ),
        migrations.AlterField(
            model_name='document',
            name='notes',
            field=models.TextField(blank=True, null=True, validators=[papermerge.core.validators.safe_character_validator], verbose_name='Notes'),
        ),
        migrations.AlterField(
            model_name='user',
            name='first_name',
            field=models.CharField(blank=True, max_length=150, verbose_name='first name'),
        ),
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(max_length=100, validators=[papermerge.core.validators.safe_character_validator], verbose_name='name'),
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('permissions', models.ManyToManyField(blank=True, to='auth.Permission', verbose_name='permissions')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='user',
            name='role',
            field=models.ForeignKey(blank=True, help_text='Determines what operations this user is authorized to perform.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.role', verbose_name='role'),
        ),
    ]



from django import template


register = template.Library()


@register.simple_tag
def document_title(document):

    return document.keywords or "Document"


@register.simple_tag
def document_url(document):

    return document.url


@register.simple_tag
def document_keywords(document):
    """
        Returns a comma separated string with document associated
        keywords. E.g. 'tax, docura, blah'
    """

    return ','.join(
        [key.name for key in document.keywords.all()]
    )



from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _


class Access(models.Model):
    # Access model guards files and folder i.e. decision
    # if a user has or hasn't permission to perform any operation
    # on give node (file/folder) is made only in respect to associated
    # access models.
    #
    # Every node (file or folder) has many access entries.
    # Every access has either one group
    # or one user associated. If a node has exactly one user
    # and one group associated - there will be two access models
    # created - one for user and one for group.

    PERM_READ = "read"
    PERM_WRITE = "write"
    PERM_DELETE = "delete"
    PERM_CHANGE_PERM = "change_perm"
    PERM_TAKE_OWNERSHIP = "take_ownership"

    ALL_PERMS = [
        PERM_READ,
        PERM_WRITE,
        PERM_DELETE,
        PERM_CHANGE_PERM,
        PERM_TAKE_OWNERSHIP
    ]

    ALLOW = "allow"
    DENY = "deny"
    MODEL_USER = "user"
    MODEL_GROUP = "group"
    OWNER_PERMS_MAP = {
        PERM_READ: True,
        PERM_WRITE: True,
        PERM_DELETE: True,
        PERM_CHANGE_PERM: True,
        PERM_TAKE_OWNERSHIP: True
    }

    node = models.ForeignKey(
        'BaseTreeNode',
        models.CASCADE,
    )
    access_type = models.CharField(
        max_length=16,
        choices=[
            (ALLOW, _('Allow')),
            (DENY, _('Deny')),
        ],
        default='allow',
    )
    # inherited access is read only
    access_inherited = models.BooleanField(
        default=False
    )
    group = models.ForeignKey(
        Group,
        verbose_name=_('group'),
        blank=True,
        null=True,
        on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        'User',
        verbose_name=_('user'),
        blank=True,
        null=True,
        on_delete=models.CASCADE
    )

    permissions = models.ManyToManyField(
        Permission
    )

    def create(node, access_inherited, access):
        if access.user:
            new_access = Access.objects.create(
                user=access.user,
                access_type=access.access_type,
                node=node,
                access_inherited=access_inherited
            )
        elif access.group:
            new_access = Access.objects.create(
                group=access.group,
                access_type=access.access_type,
                node=node,
                access_inherited=access_inherited
            )
        else:
            raise ValueError(
                "Access object must have associated either user or group"
            )

        new_access.permissions.add(
            *access.permissions.all()
        )

        return new_access

    def perms_codenames(self):
        return {p.codename for p in self.permissions.all()}

    def __str__(self):
        name = '-'
        if self.user:
            name = f"User({self.user.username})"
        elif self.group:
            name = f"Group({self.group.name})"

        typ = self.access_type

        return f"Access({name}, {typ}, ...)"

    class Meta:
        permissions = [
            ("read", "Read Access"),
            ("write", "Write Access"),
            ("delete", "Delete Access"),
            ("change_perm", "Change Permissions Access"),
            ("take_ownership", "Take Ownership Access"),
        ]

    def __hash__(self):

        name = ''

        if self.user:
            name = f"User({self.user.username})"
        elif self.group:
            name = f"Group({self.group.name})"

        typ = self.access_type
        # https://docs.python.org/3/reference/datamodel.html#object.__hash__
        return hash((name, typ))

    def __eq__(self, access):
        """
        Two access models are equal if all conditions
        are true:
        * accesses are attached to the some node
        * have same access type (allow or deny)
        * have both either same user or same group

        NOTE: inheritance flag does not matter.
        """
        if self.node.id != access.node.id:
            return False

        if self.user and access.user:
            return self.user == access.user

        if self.group and access.group:
            return self.group == access.group

        return False

    def has_perm(self, codename):
        return self.permissions.filter(
            codename=codename
        ).count() > 0

    def extract_perm_dict(self):
        result = {
            Access.PERM_READ: False,
            Access.PERM_WRITE: False,
            Access.PERM_DELETE: False,
            Access.PERM_CHANGE_PERM: False,
            Access.PERM_TAKE_OWNERSHIP: False
        }
        for perm in self.permissions.all():
            result[perm.codename] = True

        return result

    def perm_diff(self, compare_with):
        """
        passed compare with can be either:
        * a dictionary {'read': true, 'delete': false} etc
        * another Access model instance
        """
        perm_dict1 = self.extract_perm_dict()

        if isinstance(compare_with, Access):
            perm_dict2 = compare_with.extract_perm_dict()

        if isinstance(compare_with, dict):
            perm_dict2 = compare_with

        return perm_dict1 != perm_dict2

    def set_perms(self, perm_hash):

        content_type = ContentType.objects.get(
            app_label="core",
            model="access",
        )
        codenames = [
            key for key, value in perm_hash.items() if perm_hash[key]
        ]
        perms = Permission.objects.filter(
            codename__in=codenames,
            content_type=content_type
        )
        self.permissions.clear()
        if perms.count() > 0:
            self.permissions.add(*perms)

    def update_from(self, access):
        if self != access:
            return False

        self.access_type = access.access_type
        self.permissions.clear()
        self.permissions.set(
            access.permissions.all()
        )



import logging
import os

from django.db import models
from django.urls import reverse
from django.db import transaction, IntegrityError
from django.utils.translation import ugettext_lazy as _

from polymorphic_tree.managers import (
    PolymorphicMPTTModelManager,
    PolymorphicMPTTQuerySet
)

from mglib import step
from mglib.path import DocumentPath, PagePath
from mglib.pdfinfo import get_pagecount
from mglib.utils import get_assigns_after_delete

from papermerge.contrib.admin.registries import sidebar
from papermerge.core.storage import default_storage
from .kvstore import (
    KVCompNode,
    KVNode,
    get_currency_formats,
    get_date_formats,
    get_kv_types,
    get_numeric_formats
)

from .node import (
    BaseTreeNode,
    AbstractNode,
    RELATED_NAME_FMT,
    RELATED_QUERY_NAME_FMT
)
from .access import Access
from .utils import group_per_model
from .finder import default_parts_finder

from papermerge.search import index


logger = logging.getLogger(__name__)


class DocumentManager(PolymorphicMPTTModelManager):

    @transaction.atomic
    def create_document(
        self,
        user,
        title,
        lang,
        size,
        page_count,
        file_name,
        notes=None,
        parent_id=None,
        **kwargs
    ):
        """
        Creates a document.

        Special keyword argument ``parts`` is a dictionary
        of extra document parts (added by extra apps)
        """
        # extra document parts
        kw_parts = kwargs.pop('parts', {})

        parent = self._get_parent(parent_id=parent_id)

        doc = Document(
            title=title,
            size=size,
            lang=lang,
            user=user,
            parent=parent,
            notes=notes,
            file_name=file_name,
            page_count=page_count
        )
        # validate before saving
        # will raise ValidationError in case of
        # problems
        doc.full_clean()
        doc.save()
        # Important! - first document must inherit metakeys from
        # parent folder
        if parent:
            doc.inherit_kv_from(parent)

        doc.create_pages()
        doc.full_clean()

        # document parts are created regardless whether there
        # are arguments for them or no.
        self._create_doc_parts(doc, **kw_parts)
        self._create_node_parts(doc, **kw_parts)

        return doc

    def _create_node_parts(self, doc, **kw_parts):
        node_parts = default_parts_finder.find(AbstractNode)
        node_grouped_args = group_per_model(node_parts, **kw_parts)

        for model in node_parts:
            if model != BaseTreeNode:
                args = node_grouped_args.get(model, {})
                instance, _ = model.objects.get_or_create(
                    base_ptr=doc.basetreenode_ptr,
                    **args
                )
                instance.clean()

    def _create_doc_parts(self, doc, **kw_parts):
        # 1. figure out document parts
        # document_parts = [
        #    app1.Document,
        #    app2.Document,
        #    app3.Document
        # ]
        doc_parts = default_parts_finder.find(AbstractDocument)
        # 2. group arguments by document_parts
        # doc_grouped_args = {
        #    app1.Document: {},
        #    app2.Document: {},
        #    app3.Document: {}
        # }
        doc_grouped_args = group_per_model(doc_parts, **kw_parts)
        for model_klass in doc_parts:
            if model_klass != Document:
                args = doc_grouped_args.get(model_klass, {})
                instance, _ = model_klass.objects.get_or_create(
                    base_ptr=doc,
                    **args
                )
                instance.clean()

    def _get_parent(self, parent_id):
        """
        Returns parent node instance based on ``parent_id``
        """
        parent = None

        if parent_id is None or parent_id == '':
            parent = None
        else:
            try:
                parent = BaseTreeNode.objects.get(id=parent_id)
            except BaseTreeNode.DoesNotExist:
                parent = None

        return parent


class DocumentQuerySet(PolymorphicMPTTQuerySet):
    pass


class DocumentPartsManager:
    """
    Manages document parts added by extra apps.
    """

    def __init__(self, document):
        self.document = document

    def __setattr__(self, name, value):
        """
        Assign ``value`` to the attribute ``name`` on the document part model.

        document part model = django model which inherits from
        one of two classes:

            * papermerge.core.models.AbstractDocument
            * papermerge.core.models.AbstractNode

        This method provides a way to assign values to part models
        via Document model:

            doc = Document.objects.create_document(...)
            doc.parts.policy = policy  # <- this assignment is managed here
            doc.save()
        """
        model_klass = default_parts_finder.get(
            AbstractNode,
            attr_name=name
        )
        is_abs_doc = True

        if model_klass:
            is_abs_doc = False
        else:
            model_klass = default_parts_finder.get(
                AbstractDocument,
                attr_name=name
            )

        if is_abs_doc:
            # model_klass inherits from AbstractDocument
            # thus link accordingly
            base_ptr = self.document
        else:
            # model_klass inherits from AbstractNode
            # thus link accordingly
            base_ptr = self.document.basetreenode_ptr

        if model_klass:
            part_instance, _ = model_klass.objects.get_or_create(
                base_ptr=base_ptr
            )

            setattr(part_instance, name, value)
            part_instance.save()
            part_instance.clean()
        else:
            # name is a standard attribute, thus proceed with
            # standard way of assigning things
            self.__dict__[name] = value

    def __getattr__(self, name):
        """
        Looks for missing attributes in document parts
        (added by external apps).

        An example of usage:

        doc = Document.objects.get(id=3001)
        doc.parts.policy # <- this parts.policy is managed by this method.
        """
        # check for the attribute in classes that inherit from AbstractDocument
        model_klass = default_parts_finder.get(
            AbstractDocument,
            attr_name=name
        )

        # check for the attribute in classes that inherit from AbstractNode
        if not model_klass:
            model_klass = default_parts_finder.get(
                AbstractNode,
                attr_name=name
            )

        if model_klass:
            instance = model_klass.objects.get(base_ptr=self.document)
            if hasattr(instance, name):
                ret = getattr(instance, name)
                return ret


CustomDocumentManager = DocumentManager.from_queryset(DocumentQuerySet)


class Document(BaseTreeNode):

    class CannotUpload(Exception):
        pass

    #: basename + ext of uploaded file.
    #: other path details are deducted from user_id and document_id
    file_name = models.CharField(
        max_length=1024,
        default='',
    )

    notes = models.TextField(
        _('Notes'),
        blank=True,
        null=True,
    )

    size = models.BigIntegerField(
        help_text="Size of file_orig attached. Size is in Bytes",
        blank=False,
        null=False,
    )

    page_count = models.IntegerField(
        blank=False,
        default=1
    )

    # Document's version start with 0 (0 = default value)
    # Document's version is incremented everytime stapler operation
    # is applied to it (page delete, page rotate, page reorder).
    # Versioning is on file level. Means - there is no such thing as model
    # level versioning. I think this will complicate everthing just too much.
    # At any point in time, user sees/works with/searches only last version.
    version = models.IntegerField(
        blank=True,
        null=True,
        default=0
    )
    # Q: If there is no model versions, why version is introduced at all?
    # A: pdftk on every operation creates a new file... well, that new
    # file is the next version of the document.

    text = models.TextField(blank=True)

    objects = CustomDocumentManager()

    PREVIEW_HEIGHTS = (100, 300, 500)

    SMALL = PREVIEW_HEIGHTS[0]
    MEDIUM = PREVIEW_HEIGHTS[1]
    LARGE = PREVIEW_HEIGHTS[2]

    search_fields = [
        index.SearchField('title'),
        index.SearchField('text', partial_match=True, boost=2),
        index.SearchField('notes')
    ]

    def each_part(self, abstract_klasses):
        """
        Iterates through each INSTANCE of document parts which inherits
        from given ``abstract_klasses``
        """
        model_klasses = []

        for klass in abstract_klasses:
            model_klasses.extend(default_parts_finder.find(klass))

        for model_klass in model_klasses:
            # if name is an attribute of a klass which inherits
            # from AbstractDocument or AbstractNode...
            try:
                item = model_klass.objects.get(base_ptr=self)
                yield item
            except model_klass.DoesNotExist:
                # Engineering assumption:
                # There must be either 0 (zero) or 1 (one)
                # instances of model_klass with base_ptr pointing
                # to this document.
                pass

    def delete(self, *args, **kwargs):
        """
        Deletes the document and all associated parts in single transaction.

        If any of the parts objects the deletion - either by raising an
        exception or by returning 0 - deletion is aborted.

        Document part may object deletion by:

            * raising an PermissionDenied exception.
            * returning 0 count (i.e. not voting for deletion)

        Deletion function is really tricky.
        The purpose is to delete main document and its parts (satellite if you
        will) in a single database transaction. If any of the parts
        objects - the whole transaction is aborted. The problem is however,
        that if the parts that objects performs some DB operations, those
        operations will be rolled back as well (e.g. retention policy objects
        deletion, but it advances the document to the next state, or moves the
        document to another folder).
        To solve this problem (where objecting parts want to perform some DB
        operations) following approach was chosen (as if I have many options!):
        objecting parts instances are saved into a
        separate list, ``instances_which_objected`` and for those instances
        delete operation runs another round (the previous round was
        rolled back = forgotten). On second round the database
        changes are committed.
        """
        abstract_klasses = [AbstractDocument, AbstractNode]
        instance_counter = 0
        vote_counter = 0
        instances_which_objected = []

        try:
            with transaction.atomic():
                for model_instance in self.each_part(abstract_klasses):
                    instance_counter += 1

                    # Model.delete() returns a tuple.
                    # First item in tuple is number of deleted objects
                    model_instance.voting = True
                    deleted_count, _ = model_instance.delete(*args, **kwargs)
                    vote_counter += deleted_count
                    if not deleted_count:
                        instances_which_objected.append(
                            model_instance
                        )

                if vote_counter >= instance_counter:
                    # If all document parts successfully deleted
                    # then delete self
                    super().delete(*args, **kwargs)
                else:
                    raise IntegrityError
        except IntegrityError:
            # commit database changes for those parts which objected
            with transaction.atomic():
                for model_instance in instances_which_objected:
                    # a way to distinguish between two deletes
                    model_instance.voting = False
                    model_instance.delete(*args, **kwargs)

    @property
    def parts(self):
        return DocumentPartsManager(self)

    @property
    def part(self):
        return self.parts()

    def to_dict(self):
        item = {}

        first_page = None
        pages = []
        for page in self.pages.all():
            if page == self.pages.first():
                first_page = page
            pages.append(page.to_dict())

        item['id'] = self.id
        item['title'] = self.title
        item['notes'] = self.notes
        item['owner'] = self.user.username
        item['versions'] = self.get_versions()
        item['created_at'] = self.human_created_at
        item['updated_at'] = self.human_updated_at
        item['timestamp'] = self.created_at.timestamp()

        if self.parent:
            item['parent_id'] = self.parent.id
        else:
            item['parent_id'] = ''

        item['ctype'] = 'document'
        item['pages'] = pages

        tags = []
        for tag in self.tags.all():
            tags.append(tag.to_dict())
        item['tags'] = tags

        parts = []
        for sidebar_part_klass in sidebar.values():
            sidebar_part = sidebar_part_klass(self)
            parts.append(sidebar_part.to_json())

        item['parts'] = parts

        kvstore = []
        # Notice that here instead of document's metadata
        # document first page's metadata is returned.
        # This is because:
        # document's metadata == document first page's metadata
        # Why ?
        # In document viewer metadata is per page. When user
        # sees metadata in document viewer he actually sees
        # document first page metadata.
        if first_page:
            for kv in first_page.kv.all():
                kvstore.append(kv.to_dict())
        item['metadata'] = {}
        item['metadata']['kvstore'] = kvstore
        item['metadata']['currency_formats'] = get_currency_formats()
        item['metadata']['date_formats'] = get_date_formats()
        item['metadata']['numeric_formats'] = get_numeric_formats()
        item['metadata']['kv_types'] = get_kv_types()

        return item

    def assign_kv_values(self, kv_dict):
        """
        Assignes kv_dict of key value to its metadata
        and metadata of its pages.
        """
        logger.debug(
            f"assign_key_values kv_dict={kv_dict} doc_id={self.id}"
        )
        for key, value in kv_dict.items():
            # for self
            logger.debug(
                f"Assign to DOC key={key} value={value}"
            )
            self.kv[key] = value
            # and for all pages of the document
            for page in self.pages.all():
                logger.debug(
                    f"Assign to page number={page.number}"
                    f" key={key} value={value}"
                )
                try:
                    # Never (automatically) overwrite am
                    # existing Metadata value
                    if not page.kv[key]:
                        # page metadata value is empty fill it in.
                        page.kv[key] = value
                except Exception as e:
                    logging.error(
                        f"Error: page {page.number}, doc_id={self.id} has no key={key}",  # noqa
                        exc_info=e
                    )

    @property
    def kv(self):
        return KVNode(instance=self)

    def propagate_changes(
        self,
        diffs_set,
        apply_to_self,
        attr_updates=[]
    ):
        super().propagate_changes(
            diffs_set=diffs_set,
            apply_to_self=apply_to_self,
            attr_updates=attr_updates
        )
        # Access permissions are not applicable
        # for Page models, so if diffs_set contains
        # instances of Access - just return
        if (len(diffs_set)):
            first_diff = diffs_set[0]
            if len(first_diff):
                model = list(first_diff)[0]
                if isinstance(model, Access):
                    return

        # documents need to propage changes
        # to their pages

        for page in self.pages.all():
            page.propagate_changes(
                diffs_set=diffs_set,
                apply_to_self=apply_to_self,
                attr_updates=attr_updates
            )

    @property
    def kvcomp(self):
        return KVCompNode(instance=self)

    def inherit_kv_from(self, node):
        inherited_kv = [
            {
                'key': item.key,
                'kv_type': item.kv_type,
                'kv_format': item.kv_format,
                'value': item.value,
                'kv_inherited': True
            } for item in node.kv.all()
        ]
        self.kv.update(inherited_kv)

    class Meta:
        verbose_name = _("Document")
        verbose_name_plural = _("Documents")

    def __repr__(self):
        _t = self.title
        _i = self.id
        _v = self.version
        _p = self.get_pagecount()

        return f"Document(id={_i}, version={_v} title={_t}, page_count={_p})"

    def __str__(self):
        return self.title

    def get_versions(self):
        """
        Returns a list of all versions
        numbers of given document. Version
        counting starts with 0. Example:
        [0, 1, 2] - document has 3 versions.
        Original version is 0. Latest version is 2.
        """
        doc_path = self.path()
        versions_list = default_storage.get_versions(
            doc_path
        )

        return versions_list

    def is_latest_version(self, version):
        if version is None:
            return True

        version = int(version)
        return version == self.version

    def get_pagecount(self, version=None) -> int:
        """
        Returns number of pages in the document of specified version.

        There are two ways to looks at page count. One is via self.page_count
        attribute.
        That attribute reflects only the latest version's page count.
        To find out the previous version's page count, we need to have a look
        at filesystem and count number of directories of form page_<x> inside
        document's results folder of respective version.
        """
        if self.is_latest_version(version=version):
            # self.page_count attribute reflects only the latest version
            # page count
            return self.page_count

        doc_path = self.path(version=version)
        count = default_storage.get_pagecount(doc_path)

        return count

    @property
    def file_ext(self):
        _, ext = os.path.splitext(self.file_name)
        return ext

    def reorder_pages(self, new_order):

        if not isinstance(new_order, list):
            logger.error("Expecting list argument")
            return

        new_version = default_storage.reorder_pages(
            doc_path=self.path(),
            new_order=new_order
        )
        self.version = new_version
        self.save()
        self.recreate_pages()

    def delete_pages(
        self,
        page_numbers: list,
        skip_migration=False
    ):
        """
        Deletes pages with given order numbers from
        the documents.
        """
        if not isinstance(page_numbers, list):
            logger.error("Expecting list argument")
            return

        # delete pages
        new_version = default_storage.delete_pages(
            doc_path=self.path(),
            page_numbers=page_numbers,
            skip_migration=skip_migration
        )

        if new_version == self.version:
            raise Exception("Expecting version to be incremented")

        self.version = new_version

        # total pages before delete (of lastest document version)
        total_page_count = self.pages.count()
        self.pages.filter(number__in=page_numbers).delete()
        # update self.page_count attribute (for latest document version)
        self.page_count = self.pages.count()
        self.save()

        self.reassign_page_nums_after_delete(
            page_count=total_page_count,
            deleted_pages=page_numbers
        )

    def reassign_page_nums_after_delete(
        self,
        deleted_pages: list,
        page_count: int
    ):
        """
        :page_count: page count BEFORE delete operation
        """
        pairs = get_assigns_after_delete(
            total_pages=page_count,
            deleted_pages=deleted_pages
        )
        for new_version_page_num, old_version_page_num in pairs:
            page = self.pages.get(number=old_version_page_num)
            page.number = new_version_page_num
            page.save()

    def recreate_pages(self):
        """
        Recreate page models.
        """

        self.pages.all().delete()
        self.page_count = get_pagecount(
            default_storage.abspath(self.path().url())
        )
        self.save()
        self.create_pages()

    def create_pages(self):
        # Q: why doc.page_count is a valid value and yet there
        #  are no pages assigned to document model ?
        # Because page count is retrieved not via models, but by invoking
        # an external utility on pdf file!
        page_count = self.page_count

        for page_index in range(1, page_count + 1):

            preview = reverse(
                'core:preview',
                args=[self.id, 800, page_index]
            )

            page = self.pages.create(
                user=self.user,
                number=page_index,
                image=preview,
                lang=self.lang,
                page_count=page_count
            )
            page.inherit_kv_from(self)

    def update_text_field(self):
        """Update text field from associated page.text fields.

        It will update text field of all associated pages first
        (from .txt files) and then concatinate all text field
        into doc.text field.

        Returns True if document contains non empty non whitespace
        text (i.e it was OCRed)
        """
        text = ""
        txt_exists = False

        for page in self.pages.all():
            if len(page.text) == 0:
                txt_exists = page.update_text_field()
                # in case .txt for one page is not present - interrupt
                # the whole iteration
                if txt_exists:
                    page.save()
                    logger.debug(
                        f"text saved. len(page.text)=={len(page.text)}"
                    )
                else:
                    # interrupt - if one page's .txt was not found.
                    break
            else:
                logger.debug(
                    f"document_log "
                    f" username={page.user.username}"
                    f" doc_id={page.document.id}"
                    f" page_num={page.number}"
                    f" text_len={len(page.text.strip())}"
                )

            text = text + ' ' + page.text

        if txt_exists:
            # Save this document only in case all .txt
            # were present, otherwise running ./manage.py worker
            # will 'resurrect' deleted documents.
            self.text = text.strip()
            self.save()

        return len(text.strip()) != 0

    @staticmethod
    def paste_pages(
        user,
        parent_id,
        doc_pages,
        dst_document=None,
        after=False,
        before=False,
    ):
        # parent_node is an instance of BaseTreeNode
        # doc_pages is a dictionary of format:
        # {
        #    doc_id_1: [page_num_1a, page_num_2a, ...],
        #    doc_id_2: [page_num_1b, page_num_2b, ...],
        #    doc_id_3: [page_num_1c, page_num_2c, ...]
        # }
        # 1. Create a new document NEWDOC
        # 2. Build new pages for the newly created document
        # num_pages = len(doc_pages[doc_id_1]) + len(doc_pages[doc_id_2]) + ...
        # 3. for each document with ids in doc_pages.keys() (DOC):
        #     a. copy pages data from DOC to NEWDOC
        #     b. deletes pages from DOC (pages mentioned in doc_page[key] list)
        new_page_count = sum(
            [
                len(pages) for pages in doc_pages.values()
            ]
        )

        if new_page_count == 0:
            logger.warning("No pages to paste. Exiting.")
            return

        # 1. Create new document
        # 2. Build new pages for newly created document
        dst_doc_is_new = False
        if not dst_document:
            dst_document = Document.objects.create_document(
                user=user,
                parent_id=parent_id,
                lang=user.preferences['ocr__OCR_Language'],
                title="pasted.pdf",
                size=0,  # updated later, after pdftk will create new doc
                file_name="pasted.pdf",
                page_count=new_page_count
            )
            dst_doc_is_new = True

        # for each document where are pages to paste
        doc_list = []
        data_list = []
        for doc_id in doc_pages.keys():
            try:
                doc = Document.objects.get(id=doc_id)
            except Document.DoesNotExist:
                logger.warning(
                    f"While pasting, doc_id={doc_id} was not found"
                )
                return

            src = default_storage.abspath(doc.path())
            doc_path = doc.path()

            doc_list.append({'doc': doc, 'page_nums': doc_pages[doc_id]})
            data_list.append(
                {
                    'src': src,
                    'doc_path': doc_path,
                    'page_nums': doc_pages[doc_id]
                }
            )

        # returns new document version
        new_version = default_storage.paste_pages(
            dest_doc_path=dst_document.path(),
            data_list=data_list,
            dest_doc_is_new=dst_doc_is_new,
            after_page_number=after,
            before_page_number=before
        )

        dst_document.version = new_version
        dst_document.save()
        # update pages model
        dst_document.recreate_pages()

        # delete pages of source document (which where
        # cutted and pasted into new doc)
        for item in doc_list:
            item['doc'].delete_pages(
                page_numbers=item['page_nums']
            )

        return dst_document
        # TODO: update size of the new document (changed doc)

    @property
    def absfilepath(self):
        return default_storage.abspath(
            self.path().url()
        )

    def path(self, version=None):

        if version is None:
            version = self.version

        version = int(version)

        result = DocumentPath(
            user_id=self.user.id,
            document_id=self.id,
            version=version,
            file_name=self.file_name,
        )

        return result

    def page_paths(self, version=None):
        """
        Enables document instance to get quickly page
        paths:

            page_path = doc.page_path[2]
            page_path.url() # local url to second page of the doc.

        This is shortcut method when most used Step(1) is required.
        """

        results = [None]  # indexing starts from 1

        page_count = self.get_pagecount(version=version)

        for page_num in range(1, page_count + 1):
            page_path = PagePath(
                document_path=self.path(version=version),
                page_num=page_num,
                step=step.Step(1),
                page_count=self.get_pagecount(version=version)
            )
            results.append(page_path)

        return results

    def get_page_path(self, page_num, step, version=None):
        """
        For Step(1) shortcut, use doc_instance.page_eps property.
        """
        return PagePath(
            document_path=self.path(version=version),
            page_num=page_num,
            step=step,
            page_count=self.page_count
        )

    def preview_path(self, page, size=None):

        if page > self.page_count or page < 0:
            raise ValueError("Page index out of bound")

        file_name = os.path.basename(self.file_name)
        root, _ = os.path.splitext(file_name)
        page_count = self.pages_num

        if not size:
            size = "orig"

        if page_count <= 9:
            fmt_page = "{root}-page-{num:d}.{ext}"
        elif page_count > 9 and page_count < 100:
            fmt_page = "{root}-page-{num:02d}.{ext}"
        elif page_count > 100:
            fmt_page = "{root}-page-{num:003d}.{ext}"

        return os.path.join(
            self.dir_path,
            str(size),
            fmt_page.format(
                root=root, num=int(page), ext="jpg"
            )
        )

    @property
    def name(self):
        root, ext = os.path.splitext(self.file_name)
        return root

    def add_tags(self, tags):
        """
        tags is an iteratable of papermerge.core.models.Tag instances
        """
        for tag in tags:
            self.tags.add(
                tag,
                tag_kwargs={'user': self.user}
            )


class AbstractDocument(models.Model):
    """
    Common class apps need to inherit from in order
    to extend Document model.

    This class provides a way to extend papermerge.core.models.Document
    model; it does so by creating a foreign key from document extension model
    to the main model.
    These document extension models are called Document parts. They are sort
    of document model chunks holding various extra attributes.

    To add an extra attribute to main document model you need simply to
    create a new model which will inherit from AbstractDocument:

        from django.db import models
        from papermerge.core.models import AbstractDocument

        class DocumentPart(AbstractDocument):

            extra_special_id = models.CharField(
                max_length=50,
                null=True
            )

        There is one very important rule though: you should never
        instantiate DocumentPart class. The ``extra_special_id`` attribute
        will be managed via ``parts`` attribute of core document class
        for example by adding these code in core document creation signal
        handler:

            document.parts.extra_special_id = f"XYZ_{md5(file_name)}"
    """
    base_ptr = models.ForeignKey(
        Document,
        related_name=RELATED_NAME_FMT,
        related_query_name=RELATED_QUERY_NAME_FMT,
        on_delete=models.CASCADE
    )

    class Meta:
        abstract = True

    def get_pagecount(self, version=None):
        return self.base_ptr.get_pagecount(version=None)

    def get_user(self):
        return self.base_ptr.user

    def get_title(self):
        return self.base_ptr.title

    def get_file_name(self):
        return self.base_ptr.file_name

    def get_document_fields(self):
        return self.base_ptr.get_document_fields()

    def get_absfilepath(self):
        """
        Returns absolute file path of the latest version
        of the associated file.
        """
        return self.base_ptr.absfilepath


def _part_field_to_json(field_instance):

    # DocumentPart klass (which inherits from AbstractDocument) instance
    # Name of field of document part class instance
    return {
        "class": "choice",
        "value": field_instance.id,
        "choices": ((1, "one"), (2, "two"))
    }



from django.core.management.base import BaseCommand

from allauth.account.models import EmailAddress

from papermerge.core.models import User


class Command(BaseCommand):
    """
    As per allauth configurations - authentication can be configured
    to require email confirmation.
    Superuser however, is created using django's default command line utility
    ``createsuperuser``. Latter, is not aware of allauth 'mandatoryness' of
    the email field - let alone the confirmation part.
    As result, in case of mandatory email address confirmation -  superuser
    web login will always redirect in "confirmation sent".
    As workaround for such issue, this command is provided.
    User ``superuser`` command, you can:
        1. List superusers in the system.
        2. Confirm their email
    Notice that it us usual to create superuser without email address.
    Usage:
        $ ./manage.py superuser --list  # will list superuser(s)
        $ ./manage.py superuser --confirm -u admin --email admin@mail.com
    In case superuser ``admin`` does not have an email address, the email
    provided in command line, will be automatically asigned to him and then
    confirmed.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '-l',
            '--list',
            action='store_true',
            help="Lists superusers"
        )
        parser.add_argument(
            '-c',
            '--confirm',
            action='store_true',
            help="Confrim email address for given superuser"
        )
        parser.add_argument(
            '-u',
            '--username',
            help="Username of the user to confirm email for"
        )
        parser.add_argument(
            '-e',
            '--email',
            help="In case superuser does not have an email address "
            "this one will be assigned to him/her."
        )

    def handle(self, *args, **options):

        list_userusers = options.get('list', False)
        confirm = options.get('confirm', False)
        user = options.get('username', None)
        email = options.get('email', None)

        if list_userusers:
            self.list_superusers()
        elif confirm:
            self.confirm_email(user, email)

    def list_superusers(self):
        for u in User.objects.filter(is_superuser=True):
            _id = u.id
            _u = u.username
            _e = u.email
            _v = _has_verified_for_login(u)
            print(
                f"id={_id} username={_u} email={_e} verified={_v}"
            )

    def confirm_email(self, username: str, email: str):

        try:
            user = User.objects.get(
                username=username,
                is_superuser=True
            )
        except User.DoesNotExist:
            print(f"Superuser {username} was not found.")
            return

        if not (user.email or email):
            print(f"Please provide an email address.")

        if email:
            user.email = email
            user.save()

        user.confirm_email()


def _has_verified_for_login(user: User):

    email = user.email
    if email:
        ret = False
        try:
            emailaddress = EmailAddress.objects.get_for_user(user, email)
            ret = emailaddress.verified
        except EmailAddress.DoesNotExist:
            pass
    else:
        ret = EmailAddress.objects.filter(user=user, verified=True).exists()

    return ret



from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied

from papermerge.core.models import Role
from papermerge.contrib.admin.forms import RoleForm
from papermerge.contrib.admin.views import mixins as mix


class RolesView(
    LoginRequiredMixin,
    mix.RequiredPermissionMixin
):
    # only superuser can access this view
    only_superuser = True
    model = Role
    form_class = RoleForm
    success_url = reverse_lazy('admin:roles')

    def dispatch(self, request, *args, **kwargs):

        if self.request.user.is_authenticated:
            if not self._is_allowed(request):
                raise PermissionDenied()

        ret = super().dispatch(request, *args, **kwargs)

        return ret

    def _is_allowed(self, request):
        if hasattr(self, 'only_superuser'):
            if getattr(self, 'only_superuser'):
                return request.user.is_superuser

        return True


class RolesListView(
    RolesView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):
    title = _("Roles")
    required_permission = 'core.view_role'

    def get_queryset(self):
        qs = super().get_queryset()

        return qs.order_by('name')


class RoleCreateView(RolesView, generic.CreateView):

    required_permission = 'core.add_role'

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('New')
        context['action_url'] = reverse_lazy('admin:role-add')

        return context


class RoleUpdateView(RolesView, generic.UpdateView):

    required_permission = 'core.change_role'

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit')
        context['action_url'] = reverse_lazy(
            'admin:role-update',
            args=(self.object.pk,)
        )

        return context



from django.db import models
from django.contrib.auth.models import Permission


class Role(models.Model):
    """
    Named set of permissions.

    Role allows you quickly create a user of specific type i.e with
    preset permissions.
    Example: A manager role assigned to user will empower him/her to
    create, update, delete other users.
    Another example: a guest role - might have a very minimal set of
    permissions suitable only for demo mode.
    """

    name = models.CharField(max_length=64)
    permissions = models.ManyToManyField(
        Permission,
        verbose_name='permissions',
        blank=True,
    )
    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name



# Generated by Django 3.0.6 on 2020-06-02 11:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0008_auto_20200601_1812'),
    ]

    operations = [
        migrations.CreateModel(
            name='KVStore',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(blank=True, max_length=200, null=True)),
                ('namespace', models.CharField(blank=True, max_length=1024, null=True)),
                ('namespaced_key', models.CharField(blank=True, max_length=200, null=True)),
                ('value', models.TextField(blank=True, null=True)),
                ('human_key', models.CharField(blank=True, max_length=200, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='KVStorePage',
            fields=[
                ('kvstore_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.KVStore')),
                ('page_id', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Page')),
            ],
            bases=('core.kvstore',),
        ),
        migrations.CreateModel(
            name='KVStoreNode',
            fields=[
                ('kvstore_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.KVStore')),
                ('node', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.BaseTreeNode')),
            ],
            bases=('core.kvstore',),
        ),
    ]



import logging
from django.core.management.base import BaseCommand

from papermerge.contrib.admin.models import LogEntry

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """
    Display/delete logentry count per user.

    Example:

    1. Delete all (i.e. for all users) logentries older than 2 days:

    ./manage.py logentry --delete -n 2d

    2. Without any argument will display logentries per user (for all users):

    ./manage.py logentry
"""

    def add_arguments(self, parser):
        parser.add_argument(
            '-u',
            '--user',
            help="""
            Either user id or username of the user to display/delete
            logentries for. If ommited applies to all users.
        """
        )
        parser.add_argument(
            '-d',
            '--delete',
            help="Delete logentries",
            action="store_true"
        )
        parser.add_argument(
            '-n',
            '--number',
            default='7d',
            help="""
            Delete log entries older than <number><d|h|m|days|hours|minutes>.
            """
        )

    def handle(self, *args, **options):
        delete = options.get('delete')

        if delete:
            del_count, del_dict = LogEntry.objects.all().delete()
            print(f"Deleted {del_count} LogEntries.")
        else:
            count = LogEntry.objects.count()
            print(f"Total LogEntries count = {count}")



import logging

from celery import shared_task
from papermerge.core.ocr.page import ocr_page as main_ocr_page
from papermerge.core.ocr import (
    COMPLETE,
    STARTED
)

from .models import Document, Folder, Page
from .utils import Timer
from . import signal_definitions as signals

logger = logging.getLogger(__name__)


@shared_task(bind=True)
def ocr_page(
    self,
    user_id,
    document_id,
    file_name,
    page_num,
    lang,
    version=0,
    namespace=None
):
    # A task being bound (bind=True) means the first argument
    # to the task will always be the
    # task instance (self).
    # https://celery.readthedocs.io/en/latest/userguide/tasks.html#bound-tasks
    logger.info(f"task_id={self.request.id}")

    # Inform everybody interested that OCR started
    signals.page_ocr.send(
        sender='worker',
        level=logging.INFO,
        message="",
        user_id=user_id,
        document_id=document_id,
        page_num=page_num,
        lang=lang,
        namespace=namespace,
        version=version,
        status=STARTED,
    )

    with Timer() as time:

        main_ocr_page(
            user_id=user_id,
            document_id=document_id,
            file_name=file_name,
            page_num=page_num,
            lang=lang,
            namespace=namespace,
            version=version
        )

    # Inform everybody interested that OCR completed/ended
    msg = f"Ocr took {time} seconds to complete."
    signals.page_ocr.send(
        sender='worker',
        level=logging.INFO,
        message=msg,
        user_id=user_id,
        document_id=document_id,
        page_num=page_num,
        lang=lang,
        namespace=namespace,
        version=version,
        status=COMPLETE
    )

    return True


def norm_pages_from_doc(document):
    logger.debug(f"Normalizing document {document.id}")
    for page in Page.objects.filter(document=document):
        page.norm()


def norm_pages_from_folder(folder):
    for descendent in folder.get_descendants():
        if isinstance(descendent, Document):
            norm_pages_from_doc(descendent)
        elif isinstance(descendent, Folder):
            norm_pages_from_folder(descendent)
        else:
            raise ValueError("Unexpected value for descendent instance")


@shared_task
def normalize_pages(origin):
    """
    Normalize Pages. The normalization was triggered model origin.
    origin can be either a Folder or a Document
    """
    if isinstance(origin, Document):
        norm_pages_from_doc(origin)
    elif isinstance(origin, Folder):
        norm_pages_from_folder(origin)



# Generated by Django 3.0.7 on 2020-07-21 13:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0017_automate_user'),
    ]

    operations = [
        migrations.AlterField(
            model_name='automate',
            name='plugin_name',
            field=models.CharField(blank=True, choices=[], default=None, max_length=256, null=True),
        ),
    ]



import logging

from django.contrib.auth.decorators import login_required
from django.shortcuts import render, redirect
from dynamic_preferences.users.forms import user_preference_form_builder

logger = logging.getLogger(__name__)


def uniq_sections(user):
    """
    Returns number of unique instances of Sections.

    I.e. returns unique instances of
        papermerge.core.preferences.Section
    """
    preferences = user.preferences
    sections = []

    # user preferences
    for key in preferences:
        section, name = preferences.parse_lookup(key)
        pref = preferences.registry.get(section=section, name=name)
        if pref.section.visible:
            sections.append(pref.section)

    result = set(sections)
    sorted_result = sorted(
        result,
        key=lambda section: section.name
    )
    return sorted_result


@login_required
def preferences_section_view(request, section):

    Form = user_preference_form_builder(
        instance=request.user,
        section=section
    )

    if request.method == 'POST':
        form = Form(request.POST)
        if form.is_valid():
            form.update_preferences()
            return redirect('admin:preferences')

    return render(
        request,
        'admin/preferences_section.html',
        {
            'form': Form,
            'section': section
        }
    )


@login_required
def preferences_view(request):

    return render(
        request,
        'admin/preferences.html',
        {
            'sections': uniq_sections(request.user),
        }
    )



from setuptools import setup, find_namespace_packages

setup(
    packages=find_namespace_packages(
        include=['papermerge.*']
    ),
    install_requires=[
        "django == 3.1.7",
        "mglib == 1.3.8",
        "mgclipboard >= 0.3.0",
        "stapler == 1.0.0",
        "django-dynamic-preferences == 1.10.1",
        "celery >= 5.0.0",
        "django-taggit == 1.3.0",
        "imapclient",
        "djangorestframework",
        "django-rest-knox",
        "pyyaml",
        "lxml == 4.6.2",
        "django-celery-results",
        "django-allauth == 0.44.0",
        "django-polymorphic",
        "django-mptt == 0.11.0",
        "django-polymorphic-tree",
        "django-modelcluster",
        "persisting-theory",
        "django-bootstrap4",
        "python-magic"
    ]
)



import logging

from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.conf import settings

from allauth.account.signals import user_logged_in

from papermerge.core.auth import create_access
from papermerge.core.models import Access, Diff, Document, Folder, User
from papermerge.core.storage import default_storage
from papermerge.core.tasks import normalize_pages

logger = logging.getLogger(__name__)


@receiver(pre_delete, sender=Document)
def deleteFiles(sender, instance, **kwargs):
    """
    Will delete associated physical (pdf) file.

    More exactly it will delete whatever it is inside
    associated folder in which original file was saved
    (e.g. all preview images).

    Preview images as well all other associated files
    are irreversibly deleted.
    """
    logger.debug("Deleting files for doc_id={} of user_id={}".format(
        instance.id,
        instance.user.id
    ))

    default_storage.delete_doc(
        instance.path()
    )


def node_post_save(sender, node, created, *kwargs):
    if created:
        # New node instance was created.
        # Create associated Access Model:
        # node creater has full access.
        create_access(
            node=node,
            model_type=Access.MODEL_USER,
            name=node.user.username,
            access_type=Access.ALLOW,
            access_inherited=False,
            permissions=Access.OWNER_PERMS_MAP  # full access
        )

    # Consider this two persons use case:
    # User uploader uploads scans for user margaret.
    # Initially document is in uploader's Inbox folder.
    # Afterwards, uploader moves new document X into common shared_folder.
    # shared_folder has full access permissions for
    # boths uploader and margaret.
    # When margaret sees document X, she copies it into
    # her private folder X_margaret_fld. X_margaret_fld is
    # owned only by margaret.
    # Now document X's path is margaret//X_margaret_fld/X.pdf
    # If X.pdf access permissions stay same, then uploader will
    # still have access to X.pdf (via search) which means,
    # that margaret will need to change manually X.pdf's
    # access permissions. To avoid manual change of access
    # permissions from margaret side - papermerge feature
    # is that X.pdf inherits access permissions from new
    # parent (X_margaret_fld). Thus, just by copying it,
    # X.pdf becomes margaret private doc - and uploader
    # lose its access to it.
    if node.parent:  # current node has a parent?
        # Following statement covers case when node
        # is moved from one parent to another parent.

        # When node moved from one parent to another
        # it get all its access replaced by access list of the
        # parent
        access_diff = Diff(
            operation=Diff.REPLACE,
            instances_set=node.parent.access_set.all()
        )
        node.propagate_changes(
            diffs_set=[access_diff],
            apply_to_self=True
        )
    else:
        # In case node has no parent, all its access permission
        # remain the same.
        pass


@receiver(post_save, sender=Folder)
def save_node_folder(sender, instance, created, **kwargs):
    node_post_save(sender, instance, created, kwargs)


@receiver(post_save, sender=Document)
def save_node_doc(sender, instance, created, **kwargs):
    node_post_save(sender, instance, created, kwargs)


@receiver(post_save, sender=Document)
def inherit_metadata_keys(sender, instance, created, **kwargs):
    """
    When moved into new folder, documents will inherit their parent
    metadata keys
    """
    # if doc has a parent
    if instance.parent:
        instance.inherit_kv_from(instance.parent)
        for page in instance.pages.all():
            page.inherit_kv_from(instance.parent)
    else:
        for page in instance.pages.all():
            page.inherit_kv_from(instance)


@receiver(post_save, sender=Folder)
def inherit_metadata_keys_from_parent(sender, instance, created, **kwargs):
    """
    When created or moved folders will inherit metadata keys from their
    parent.
    """
    # if folder was just created and has a parent
    if created and instance.parent:
        instance.inherit_kv_from(instance.parent)


@receiver(post_save, sender=Document)
def normalize_pages_from_doc_handler(sender, instance, created, **kwargs):
    """Update async this document pages' attributes page.norm_*
    """
    logger.debug("Post save doc => normalize_pages")
    normalize_pages(origin=instance)


@receiver(post_save, sender=Folder)
def normalize_pages_from_folder_handler(sender, instance, created, **kwargs):
    """Update async folder pages attributes page.norm_*
    """
    normalize_pages(origin=instance)


def _user_init(user):
    """
    Create user specific data:
        1. Inbox folder
    """
    if settings.PAPERMERGE_CREATE_SPECIAL_FOLDERS:
        Folder.objects.get_or_create(
            title=Folder.INBOX_NAME,
            parent=None,
            user=user
        )


@receiver(post_save, sender=User)
def user_init(sender, instance, created, **kwargs):
    """
    Signal sent when user model is saved
    (create=True if user was actually created).
    Create user specific data when user is initially created
    """
    if created:
        _user_init(user=instance)


@receiver(user_logged_in)
def user_logged_in_handler(sender, request, user, **kwargs):
    """
    Signal sent when user logs in.
    Just double check is user specific data is there.
    Spares admin for logging additional management scripts.
    """
    _user_init(user=user)



import logging

from django.conf import settings
from django.core.management.base import BaseCommand
from papermerge.core.importers.imap import (
    import_attachment,
    email_iterator,
    select_inbox
)
from papermerge.core.importers.imap import login as imap_login


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = """Helper command to troubleshoot IMAP account.

    It can:

        * check if Papermerge is able to connect to IMAP account
        * count number of UNSEEN email messages
        * fetch/import documents from IMAP account
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '-c',
            '--connect',
            action='store_true',
            help="Check if can connect to IMAP account"
        )
        parser.add_argument(
            '--count',
            action='store_true',
            help="Counts total number of UNSEEN email messages"
            " with attachments. Note that running this command will "
            "mark messages as SEEN. Suitable only for "
            "troubleshooting imap importer."
        )
        parser.add_argument(
            '--import',
            action='store_true',
            help="Imports documents from email attachments. "
            "Does not run in a loop i.e. imports currently 'UNSEED'"
            " messages and quit."
        )

    def handle(self, *args, **options):

        connect_action = options.get('connect', False)
        count_action = options.get('count', False)
        import_action = options.get('import', False)

        # options used across all actions (i.e. connect, count, import)
        self._imap_server = settings.PAPERMERGE_IMPORT_MAIL_HOST
        self._username = settings.PAPERMERGE_IMPORT_MAIL_USER
        self._password = settings.PAPERMERGE_IMPORT_MAIL_PASS
        self._delete = settings.PAPERMERGE_IMPORT_MAIL_DELETE
        self._by_user = settings.PAPERMERGE_IMPORT_MAIL_BY_USER
        self._by_secret = settings.PAPERMERGE_IMPORT_MAIL_BY_SECRET
        self._inbox_name = settings.PAPERMERGE_IMPORT_MAIL_INBOX

        if connect_action:
            self._connect_action()
        elif count_action:
            self._count_action()
        elif import_action:
            self._import_action()

    def _connect_action(self):
        if self._all_credentials_provided():
            server = imap_login(
                imap_server=self._imap_server,
                username=self._username,
                password=self._password,
            )
            self.stdout.write("Connection to IMAP server: ", ending='')
            if server:
                self.stdout.write(self.style.SUCCESS("OK"))
            else:
                self.stdout.write(self.style.ERROR("FAIL"))

    def _count_action(self):
        server = imap_login(
            imap_server=self._imap_server,
            username=self._username,
            password=self._password,
        )
        if server:
            server = select_inbox(server, self._inbox_name, readonly=True)
            emails_count = list(email_iterator(server))

            self.stdout.write(f"Total count: {len(emails_count)}")

    def _import_action(self):

        if self._all_credentials_provided():
            import_attachment(
                imap_server=self._imap_server,
                username=self._username,
                password=self._password,
                delete=self._delete,
                inbox_name=self._inbox_name,
                by_user=self._by_user,
                by_secret=self._by_secret
            )
        else:
            self.stdout.write(
                self.style.ERROR("Not all credentials are provided.")
            )

    def _all_credentials_provided(self):
        """
        Returns True only if all IMAP host, username and password
        are non-empty.
        Returns False if at least one of imap host, username or password
        is empty.
        """

        if not self._imap_server:
            self.stdout.write(
                self.style.ERROR("IMPORT_MAIL_HOST is empty.")
            )
            return False

        if not self._username:
            self.stdout.write(
                self.style.ERROR("IMPORT_MAIL_USER is empty.")
            )
            return False

        if not self._password:
            self.stdout.write(
                self.style.ERROR("IMPORT_MAIL_PASS is empty.")
            )
            return False

        return True



# Generated by Django 3.1.3 on 2021-02-16 09:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0038_auto_20210205_0915'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='user',
            name='current_storage_size',
        ),
    ]



import ssl
import email
import logging

from imapclient import IMAPClient
from imapclient.exceptions import LoginError
from imapclient.response_types import BodyData
from email.message import EmailMessage

from django.db.models import Q

from papermerge.core.import_pipeline import IMAP, go_through_pipelines
from papermerge.core.models import User


logger = logging.getLogger(__name__)


def login(imap_server, username, password):
    """
    Login to provided IMAP server.

    All arguments :imap_server:, :username: and :password:
    are string instances.

    On successful login returns a non empty ``imapclient.IMAPClient`` instance.
    """

    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE

    imap_client = IMAPClient(
        imap_server,
        ssl_context=ssl_context
    )

    try:
        imap_client.login(username, password)
    except LoginError:
        logger.error(
            "IMAP Import: ERROR. Login failed."
        )
        return None

    return imap_client


def trigger_document_pipeline(
    email_message: EmailMessage,
    user=None,
    skip_ocr=False
):
    """
    email_message is instance of ``email.message.EmailMessage``.
    Where email package is python standard (for python >= 3.6)
    library for managing email messages.
    """
    ingested = False

    if not isinstance(email_message, EmailMessage):
        raise ValueError("Expecting email.message.EmailMessage instance")

    for part in email_message.iter_attachments():
        try:
            payload = part.get_content()
        except KeyError:
            continue
        init_kwargs = {'payload': payload, 'processor': IMAP}
        apply_kwargs = {
            'user': user,
            'name': part.get_filename(),
            'skip_ocr': skip_ocr
        }
        doc = go_through_pipelines(init_kwargs, apply_kwargs)
        if doc is not None and not ingested:
            ingested = True

    return ingested


def contains_attachments(structure: BodyData):
    """
    Looks in email's BodyData for attachments.

    Returns True if BodyData instance has an attachment.

    BodyData is basically a list of tuples (list of parts) which in their turn
    can be nested as well (each part might be a part as well). We traverse
    recursively the bodystructure until the parts are not lists anymore(not
    multiparts).
    """
    if not isinstance(structure, BodyData):
        return False

    if structure.is_multipart:
        for part in structure:
            if isinstance(part, list):
                for element in part:
                    if contains_attachments(element):
                        return True

    # Here we have bodydata which is not multpart.
    # Here BodyData is a tuple.
    # Look for b'attachment' tuple's elements
    # This tuple can still have other tuples.
    # NOTE: Comparing lowercase here is important.
    # Some IMAP servers will return b'ATTACHMENT'. Thus we
    # need to lowercase it first.
    for it in structure:
        if isinstance(it, tuple) and len(it) > 0:
            if hasattr(it[0], 'lower') and it[0].lower() == b'attachment':
                return True
        # item is a scalar here
        # is this scalar b'attachment' ?
        elif hasattr(it, 'lower') and it.lower() == b'attachment':
            return True

    return False


def get_secret(str_or_list):
    """
    Find and returns mail secret.

    Mail secret uses following format:

        SECRET{ some-secret }

    note that there are no spaces between all capitalized keyword
    SECRET and immediately following it curly brackets.
    However, the text inside curly brackets can be surrounded by white
    spaces. In example above ``get_secret``
    function returns "some-secret" string.

    :str_or_list: argument can be either a string of a list of
    strings.
    """

    text = str_or_list

    if isinstance(str_or_list, list):
        text = "\n".join(str_or_list)

    try:
        message_secret = text.split('SECRET{')[1].split('}')[0]
    except IndexError:
        message_secret = None

    if message_secret:
        message_secret = message_secret.strip()

    return message_secret


def match_by_user(to_field, from_field):
    """
    Returns first user with email address matching to_field or from_field.

    Note that search is perfomed only on users with ``mail_by_user`` attribute
    set to True.
    :user: is instance of ``papermerge.core.models.User``
    :to_field: and :from_field: are email addresses as string.

    If no user matches - returns None.
    """

    user_found = User.objects.filter(
        Q(email=from_field) | Q(email=to_field)
    ).first()

    if user_found and user_found.preferences['email_routing__by_user']:
        logger.debug(
            f"{IMAP} importer: found user {user_found} from email"
        )
        return user_found

    return None


def match_by_secret(message_secret):
    """
    Returns first user with matching mail_secret.

    :user: is instance of ``papermerge.core.models.User``
    Search is performed only on users with ``mail_by_secret`` flag set to
    True.
    If no user matches - returns None.
    """
    user_found = User.objects.filter(
        userpreferencemodel__section="email_routing",
        userpreferencemodel__name="mail_secret",
        userpreferencemodel__raw_value=message_secret,
    ).first()

    if user_found and user_found.preferences['email_routing__by_secret']:
        logger.debug(
            f"{IMAP} importer: found user {user_found} from secret"
        )
        return user_found

    return None


def get_matching_user(
    email_message: EmailMessage,
    by_user=False,
    by_secret=False
):
    """
    Returns ``papermerge.core.models.User`` instance
    of the user for whom given email message is addressed.

    email_message is instance of ``email.message.EmailMessage``.
    Where email package is python standard (for python >= 3.6)
    library for managing email messages.

    Will return None when no user matched.
    """

    body_text = None
    user_found = None

    if not isinstance(email_message, EmailMessage):
        raise ValueError("Expecting email.message.EmailMessage instance")

    from_field = email.utils.parseaddr(
        email_message.get('From'))[1]

    to_field = email.utils.parseaddr(
        email_message.get('To'))[1]

    body = email_message.get_body()

    if body is not None:
        body_text = body.as_string()

    subject_text = email_message.get('Subject')

    # secret can be in body or in subject
    message_secret = get_secret(
        [body_text, subject_text]
    )

    if by_user:
        user_found = match_by_user(
            to_field=to_field,
            from_field=from_field
        )

    # matched by_user?
    if user_found:
        # yes, it matched, just return.
        return user_found

    if by_secret and message_secret is not None:
        user_found = match_by_secret(message_secret)

    return user_found


def select_inbox(
    imap_client,
    inbox_name,
    readonly=False
):
    """
    Thin layer over imap_client.select_folder function
    """
    if not isinstance(imap_client, IMAPClient):
        raise ValueError("Expecting IMAPClient instance as first argument")

    try:
        imap_client.select_folder(
            inbox_name,
            readonly=False
        )
    except Exception:
        logger.error(
            f"IMAP import: Failed to select folder. "
            f"Maybe user needs read write access to the folder "
            f"\"{inbox_name}\"?"
        )
        return

    return imap_client


def email_iterator(
    imap_client: IMAPClient,
    delete=False
):
    """
    Generator used for lazy iteration over
    all UNEED email massages WITH attachment.

    Yields ``email.message.EmailMessage`` instances.
    Where email package is python standard (for python >= 3.6)
    library for managing email messages.
    """

    if not isinstance(imap_client, IMAPClient):
        raise ValueError("Expecting IMAPClient instance as first argument")

    messages = imap_client.search(['UNSEEN'])

    logger.debug(
        f"IMAP Import: UNSEEN messages {len(messages)} count"
    )

    messages_structure = imap_client.fetch(messages, ['BODYSTRUCTURE'])

    for uid, structure in messages_structure.items():
        if not contains_attachments(structure[b'BODYSTRUCTURE']):
            messages.remove(uid)

    for uid, message_data in imap_client.fetch(messages, ['RFC822']).items():
        body = message_data[b'RFC822']
        email_message = email.message_from_bytes(
            body,
            policy=email.policy.default
        )
        # mark e-mail as read/seen, so that it won't be processed or imported
        # multiple times
        imap_client.add_flags(uid, br'\Seen')

        yield email_message

    if delete:
        # flag all processed mails with \Deleted
        imap_client.delete_messages(messages)
        # remove all messages from the currently selected folder that have the
        # \Deleted flag set
        imap_client.expunge()


def import_attachment(
    imap_server: str,
    username: str,
    password: str,
    delete=False,
    inbox_name="INBOX",
    by_user=False,
    by_secret=False
):
    imap_client = login(
        imap_server=imap_server,
        username=username,
        password=password
    )

    if not imap_client:
        logger.info(
            f"IMAP import: Failed to login to imap server {imap_server}."
            " Please double check IMAP account credentials."
        )
        return

    imap_client = select_inbox(imap_client, inbox_name)

    for email_message in email_iterator(imap_client, delete=delete):
        user = get_matching_user(
            email_message,
            by_user=by_user,
            by_secret=by_secret
        )
        trigger_document_pipeline(
            email_message=email_message,
            user=user
        )
